Unit D64;

{ Added v2.3 :

  SAVE "@0:FILENAME" to replace a file

  LOAD "$0:A?B*" - support pattern on load directory

  OPEN 15,8,15,COMMAND$ / or PRINT#15,COMMAND$

  - support NEW,RENAME,SCRATCH

  - NEW create a new .D64 !

}

Interface

{-$DEFINE DEBUG - do not use StatusBar & GetKey with DBG64 !!!! }

uses
 fdos,dialogs,keyboard,dos,data,strings;

Type
{ Block Availability Map }
 TBAM=record
  Track :byte; { Directory }
  Sector:byte;
  Format:char; { 'A' }
  Zero  :byte; {  0  }
  SAT   :array[1..35] of longint; { Sector Allocation Table }
  Name  :array[1..18] of char;    { Label +  (160=RVS+32) }
  ID    :word; { '00'    }
  Space :byte; { ' '+128 }
  Chars :word; { '2A'    }
  Spaces:array[1..4] of byte;  { ' '+128 }
  Zeros :array[1..85] of byte; { 0 }
 end;

{ Directory entry structure }
 TDirEntry=record
  Kind   :byte;
  nTrack :byte;
  nSector:byte;
  Name   :array[1..16] of char;
  sTrack :byte;
  sSector:byte;
  Len    :byte;
  pad0   :longint;
  oTrack :byte;
  oSector:byte;
  Blocks :word;
  pad1   :word;
 end;

{ Directory block structure }
 TDirectory=record
  NextTrack :byte;
  NextSector:byte;
  Entry     :array[0..7] of TDirEntry;
 end;

 TSector=record
  NextTrack,NextSector:byte;
  Data:array[0..253] of byte;
 end;

function OpenDevice:boolean;
procedure CloseDevice;

function Load(Channel:byte; AFileName:string):byte;
function ReadByte(Channel:byte):byte;

function Save(Channel:byte; AFileName:string):byte;
function WriteByte(Channel,Value:byte; EndOfFile:boolean):byte;

Procedure ExecCommand;

Var
 D64File:string;

 Command:string;

Implementation

uses iec;

const
 SectorCount:array[1..40] of byte=(
  21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,21,
  19,19,19,19,19,19,19,
  18,18,18,18,18,18,
  17,17,17,17,17,
  17,17,17,17,17	{ Tracks 36..40 }
 );

 SectorOfs:array[1..40] of integer=(
  0,21,42,63,84,105,126,147,168,189,210,231,252,273,294,315,336,
  357,376,395,414,433,452,471,
  490,508,526,544,562,580,
  598,615,632,649,666,
  683,700,717,734,751	{ Tracks 36..40 }
 );

 KindName:array[0..15] of array[1..3] of char=(
  'DEL','SEQ','PRG','USR','REL','EL?','EQ?','RG?','SR?','EL?','L??',
  'Q??','G??','R??','L??','???'
 );

 ErrorMessage:array[0..11] of Pchar=(
  '00, OK',
  '25,WRITE ERROR',
  '26,WRITE PROTECT ON',
  '30,SYNTAX ERROR',
  '33,SYNTAX ERROR',
  '60,WRITE FILE OPEN',
  '61,FILE NOT OPEN',
  '62,FILE NOT FOUND',
  '67,ILLEGAL TRACK OR SECTOR',
  '70,NO CHANNEL',
  '73,CBM DOS V2.6 1541',
  '74,DRIVE NOT READY'
 );

 errNone     =0;
 errWrite    =1;
 errReadOnly =2;
 errUnknow   =3;
 errSyntax   =4;
 errWriteOpen=5;
 errNotOpen  =6;
 errNotFound =7;
 errInvalidTS=8;
 errNoChannel=9;
 errVersion  =10;
 errNotReady =11;


Var
 Handle:word;

 HeaderSize:word;

 BAM:tBAM;
 FreeBlocks:word;

 FileName:string;

 Channels:array[0..15] of Record
  SectorSize:word;
  SectorIndex:word;
  FillSector:(fsNone,fsRead,fsDir,fsDone);
  Sector:TSector;
 end;

 Dir:TDirectory;
 DirTrack :word;
 DirSector:word;
 DirEntry :byte;

(*
 Sector     :TSector;
 SectorSize :word;
 SectorIndex:word;
 NextSector :word;
 NextTrack  :word;
 FillSector :(fsNone,fsRead,fsDir,fsDone,fsBlock);
*)

procedure SetError(ErrCode,ErrTrack,ErrSector:word);
 begin
  With Channels[15] do begin
   SectorSize:=StrLen(ErrorMessage[ErrCode]);
   SectorIndex:=0;
   FillSector:=fsDone;
   Move(ErrorMessage[ErrCode]^,Sector.Data,SectorSize);
   Sector.Data[SectorSize]:=ord(','); inc(SectorSize);
   Sector.Data[SectorSize]:=ord('0')+ErrTrack div 10; Inc(SectorSize);
   Sector.Data[SectorSize]:=ord('0')+ErrTrack mod 10; Inc(SectorSize);
   Sector.Data[SectorSize]:=ord(','); inc(SectorSize);
   Sector.Data[SectorSize]:=ord('0')+ErrSector div 10; Inc(SectorSize);
   Sector.Data[SectorSize]:=ord('0')+ErrSector mod 10; Inc(SectorSize);
   Sector.Data[SectorSize]:=13; inc(SectorSize);
  end;
 end;



Function SectorOffset(Track,Sector:word):longint;
 begin
  if (Track<1)or(Track>40)
   or(Sector>=SectorCount[Track]) then
    SectorOffset:=-1
   else
    SectorOffset:=(SectorOfs[Track]+Sector) shl 8+HeaderSize;
 end;

{ Read sector (256 bytes) }
function ReadSector(Track,Sector:byte; Var Data):boolean;
 var
  offset:longint;
 begin
{  StatusBar('Reading Sector'); }
  Offset:=SectorOffset(Track,Sector);
  ReadSector:=(Offset>=0)
           and(SeekFile(Handle,Offset))
           and(ReadFile(Handle,Data,256)=256);
 end;

function WriteSector(Track,Sector:byte; Var Data):boolean;
 var
  offset:longint;
 begin
{  StatusBar('Writing Sector'); }
  Offset:=SectorOffset(Track,Sector);
  WriteSector:=(Offset>=0)
            and(SeekFile(Handle,Offset))
            and(WriteFile(Handle,Data,256)=256);
 end;

Function AllocSector(Var ATrack,ASector:byte):boolean;
 Var
  T,S:byte;
  F:longint;
  A:longint;
 begin
  If FreeBlocks=0 then begin
{   StatusBar('Disk full'); }
   AllocSector:=False;
   exit;
  end;
  For T:=1 to 35 do begin { Search in all Tracks }
   F:=BAM.SAT[T]; { SAT info for this Track }
   if (F AND $FF)>0 then begin { Any Sector left ? }
    A:=$100; { Sector 0 flag }
    S:=0;
    While (F AND A)=0 do begin { While Sector allready used }
     A:=A SHL 1; { next Sector flag }
     INC(S);     { next Sector }
    end;
    ATrack :=T;
    ASector:=S;
    Dec(BAM.SAT[T],A+1); { Clear Sector Flag & Dec Number of tracks left }
    Dec(FreeBlocks);
    AllocSector:=True;
    exit;
   end;
  end;
  AllocSector:=False;
 end;

Procedure CreateDevice(Const AName:string;AID:word);
 var
  t,s:word;
  a:longint;
  Sector:TSector;
 begin
  Handle:=CreateFile(D64File);
  if Handle=0 then exit;
  FillChar(Sector,SizeOf(Sector),1);
  Sector.NextTrack:=75;
  for t:=1 to 17 do begin
   for s:=1 to SectorCount[t] do WriteFile(Handle,Sector,SizeOf(Sector));
  end;
  with BAM do begin
   Track :=18;
   Sector:=1;
   Format:='A';
   Zero  := 0 ;
   for t:=1 to 35 do begin
    s:=SectorCount[t];
    a:=0;
    while s>0 do begin
     a:=a shl 1+$100;
     dec(s);
    end;
    inc(a,SectorCount[t]);
    SAT[t]:=a;
   end;
   dec(SAT[18],$302); { sector 1 & 2 are used }
   FillChar(Name,SizeOf(Name),160);
   for s:=1 to length(AName) do Name[s]:=AName[s];
   ID:=AID;
   Space:=32+128;
   Chars:=ord('2')+ord('A')*256;
   FillChar(Spaces,SizeOf(Spaces),32+128);
   FillChar(Zeros,SizeOf(Zeros),0);
  end;
  WriteFile(Handle,BAM,SizeOf(BAM));
  FillChar(Dir,SizeOf(DIR),0);
  Dir.NextSector:=255;
  WriteFile(Handle,Dir,SizeOf(Dir)-2);
  for s:=3 to SectorCount[18] do writeFile(Handle,Sector,SizeOf(Sector));
  for t:=19 to 35 do begin
   for s:=1 to SectorCount[t] do WriteFile(Handle,Sector,SizeOf(Sector));
  end;
 end;

function OpenDevice:boolean;
 var
  magic:longint;

  function ReadBAM:boolean;
   var
    i:byte;
   begin
    if (ReadFile(Handle,Magic,4)=4) and (Magic=$64411543) then
     HeaderSize:=64
    else
     HeaderSize:=0;

    if ReadSector(18,0,BAM) then begin
     ReadBAM:=True;
     for i:=1 to 18 do if BAM.Name[i]=#$A0 then BAM.Name[i]:=' ';
     FreeBlocks:=0;
     for i:=1 to 35 do inc(FreeBlocks,BAM.SAT[i] and $FF);
    end else
     ReadBAM:=False;
   end;

 begin
  if Handle=0 then begin
  {$IFNDEF DEBUG}
   if D64File='' then SelectFile('Select a D64 File','D64',D64File);
  {$ENDIF}
   if D64File='' then Handle:=0 else Handle:=OpenFile(D64File,fmReadWrite);
   if (Handle>0) then begin
    if (FileSize(Handle)<683*256)or(not ReadBAM) then begin
     CloseFile(Handle);
     Handle :=0;
    end;
   end;
  end;
  OpenDevice:=(Handle>0);
 end;

Procedure CloseDevice;
 begin
  if handle>0 then begin
   CloseFile(Handle);
   Handle:=0;
  end;
 end;

function LoadDirectory:byte; { $0:A?* }

 var
  e:word;
  i:byte;
  q:char;

  FirstLine:record
   ADDR :word;
   Link :word;
   Drive:word;
   RVSOn:byte;
   Quot1:char;
   Name :array[1..16] of char;
   Quot2:char;
   Spc1 :char;
   ID   :word;
   Spc2 :char;
   CHARS:word;
   ZERO :byte;

  end;

 begin
  { $0:A?* }
  if Copy(FileName,1,3)='$0:' then begin
   Delete(FileName,1,3);
   While Length(FileName)<16 do FileName:=FileName+'';
  end else begin
   FileName:='*';
  end;

  { Directory title }
  with FirstLine do begin
   ADDR :=$0401;
   Link :=$0101;
   Drive:=$0000;
   RVSOn:=$12;
   Quot1:='"';
   Move(BAM.Name,Name,16);
   Spc1 :=' ';
   Quot2:='"';
   ID   :=BAM.Id;
   Spc2 :=' ';
   CHARS:=BAM.Chars;
   ZERO :=0;
  end;

  with channels[0] do begin
   Move(FirstLine,Sector.Data,SizeOf(FirstLine));
   SectorSize:=SizeOf(FirstLine);
   SectorIndex:=0;
   FillSector:=fsDir;
  end;

  Dir.NextTrack :=BAM.Track;
  Dir.NextSector:=BAM.Sector;

  LoadDirectory:=ST_OK;

 end;

Procedure NextDirectory;
 var
  e,i:word;
  q:char;

  NewLine:record
   Link  :word;
   Blocks:word;
   Space1:char;
   Space2:char;
   Space3:char;
  end;

  EntryLine:record
   Quote1:char;
   Name  :array[1..16] of char;
   Quote2:char;
   Flag1 :char;
   Kind  :array[1..3] of char;
   Flag2 :char;
   ZERO  :byte;
  end;

  LastLine:record
   Link:word;
   Free:word;
   Text:array[1..12] of char;
   ZERO1:word;
   ZERO2:word;
  end;

  match:boolean;

 begin
  with Channels[0] do begin

   SectorSize :=0;
   SectorIndex:=0;

   While(SectorSize=0)
   and (Dir.NextTrack>0)
   and (ReadSector(Dir.NextTrack,Dir.NextSector,Dir)) do begin
    { Scan all 8 entries of a block }
    for e:=0 to 7 do with Dir.Entry[e] do if Kind<>0 then begin

    { Now support pattern :) }
     match:=True;
     for i:=1 to length(FileName) do begin
      if FileName[i]='*' then break; { Match }
      if (FileName[i]<>'?')and(FileName[i]<>Name[i]) then begin
       match:=false;
       break;
      end;
     end;

     IF MATCH then begin

      NewLine.Link  :=$0101;
      NewLine.Blocks:=Blocks;
      NewLine.Space1:=' ';
      NewLine.Space2:=' ';
      NewLine.Space3:=' ';

      i:=SizeOf(NewLine);
      if Blocks>=10  then dec(i);
      if Blocks>=100 then dec(i);

      Move(NewLine,Sector.Data[SectorSize],i);
      inc(SectorSize,i);

      EntryLine.Quote1:='"'; q:='"';
      for i:=1 to SizeOf(Name) do
       if Name[i]='' then begin Name[i]:=q; q:=' '; end;

      Move(Name,EntryLine.Name,SizeOf(Name));
      EntryLine.Quote2:=q;

      if Kind and $80=0 then EntryLine.Flag1:='*' else EntryLine.Flag1:=' ';
      Move(KindName[Kind and $F],EntryLine.Kind,3);
      if Kind and $40=0 then EntryLine.Flag2:=' ' else EntryLine.Flag2:='<';

      EntryLine.Zero  :=0;

      Move(EntryLine,Sector.Data[SectorSize],SizeOf(EntryLine));
      Inc(SectorSize,SizeOf(EntryLine));
     end;

    end; { for entry }

  end;

  if SectorSize=0 then begin
   with LastLine do begin
    Link:=$0101;
    Free:=FreeBlocks;
    Text:='BLOCKS FREE.';
    ZERO1:=0;
    ZERO2:=0;
   end;
   Move(LastLine,Sector.Data,SizeOf(LastLine));
   SectorSize:=SizeOf(LastLine);
   FillSector:=fsDone;
  end;

  end;
 end;


Function FindFile(FileName:string; Var ATrack,ASector:byte;Exact:boolean):boolean;
 var
    n:integer;
  found:boolean;

 begin

  Dir.NextTrack :=BAM.Track;
  Dir.NextSector:=BAM.Sector;

  DirTrack :=BAM.Track;
  DirSector:=BAM.Sector;

  While Length(FileName)<16 do FileName:=FileName+'';

  while(Dir.NextTrack<>0)
   and (ReadSector(Dir.NextTrack,Dir.NextSector,Dir)) do begin
    { Scan all 8 entries of a block }
    for DirEntry:=0 to 7 do with Dir.Entry[DirEntry] do if (Kind and $F)>0 then begin
     found:=true;
     If Exact then begin { SAVE }
      for n:=1 to Length(FileName) do
       if FileName[n]<>Name[n] then found:=false;
     end else begin { LOAD }
      for n:=1 to Length(FileName) do begin
       if FileName[n]='*' then break; { Found }
       if (FileName[n]<>'?')and(FileName[n]<>Name[n]) then begin
        found:=false;
        break;
       end;
      end;
     {
      if FileName<>'*' then
       for n:=1 to Length(FileName) do if FileName[n]<>Name[n] then found:=false;
     }
     end;
     if found then begin
      ATrack  :=nTrack;
      ASector :=nSector;
      FindFile:=True;
      exit;
     end;
    end;
    DirTrack :=Dir.NextTrack;
    DirSector:=Dir.NextSector;
   end;
  Findfile:=False;

 end;

Function ReadNextSector(Channel:byte):byte;
 begin
  with Channels[Channel] do begin
  if ReadSector(Sector.NextTrack,Sector.NextSector,Sector) then begin
   if Sector.NextTrack=0 then
    SectorSize:=Sector.NextSector-1
   else
    SectorSize:=254;
   SectorIndex:=0;
   if Sector.NextTrack=0 then FillSector:=fsDone else FillSector :=fsRead;
   ReadNextSector:=ST_OK;
  end else begin
   {$ifndef debug}
    StatusBar('NextSector error');
   {$endif}
   SectorSize:=0;
   FillSector:=fsNone;
   ReadNextSector:=ST_FILENOTFOUND;
  end;
  end;
 end;

function LoadFile(Channel:byte):byte;
 var
  p:byte;
 begin
  { Convert FileName ??? }
  p:=pos(':',FileName);
  if p>0 then begin
   delete(FileName,1,p);
  end;

  p:=pos(',',FileName);
  if p>0 then begin
   FileName[0]:=chr(p-1);
  end;

  with Channels[Channel] do begin

  { Now go on }
  if FindFile(FileName,Sector.NextTrack,Sector.NextSector,False) then
   LoadFile:=ReadNextSector(Channel)
  else begin

  {$IFNDEF DEBUG}
   StatusBar('['+FileName+'] not found');
   GetKey;
  {$ENDIF}

   SetError(errNotFound,0,0);

   LoadFile   :=62{ST_NOTPRESENT{TIMEOUT};
   SectorSize :=0;
   SectorIndex:=0;
   FillSector :=fsNone;
  end;

  end;
 end;

Function NewFile(Var ATrack,ASector:byte):byte;
 var
  n:byte;
  found:boolean;
 begin
  Dir.NextTrack :=BAM.Track;
  Dir.NextSector:=BAM.Sector;

  DirTrack :=BAM.Track;
  DirSector:=BAM.Sector;

  while(Dir.NextTrack<>0)
   and (ReadSector(Dir.NextTrack,Dir.NextSector,Dir)) do begin
    { Scan all 8 entries of a block }
    for DirEntry:=0 to 7 do with Dir.Entry[DirEntry] do if Kind=0 then begin

     Kind   :=$82;
     AllocSector(nTrack,nSector);
     For n:=1 to 16 do
      if n<=Length(FileName) then Name[n]:=FileName[n] else Name[n]:='';
     sTrack :=0;
     sSector:=0;
     Len    :=0;
     pad0   :=0;
     oTrack :=0;
     oSector:=0;
     Blocks :=0;
     pad1   :=0;

     ATrack :=nTrack;
     ASector:=nSector;

     NewFile:=ST_OK;
     exit;

    end;

    if DirTrack>0 then begin
     DirTrack :=Dir.NextTrack;
     DirSector:=Dir.NextSector;
    end;
   end;

 { Dir.NextTrack=0 }
  AllocSector(Dir.NextTrack,Dir.NextSector);
  WriteSector(DirTrack,DirSector,Dir);

  DirTrack :=Dir.NextTrack;
  DirSector:=Dir.NextSector;

  { New Directory }
  FillChar(Dir,SizeOf(Dir),0);
  Dir.NextSector:=255;
  For DirEntry:=1 downto 7 do begin
   FillChar(Dir.Entry[DirEntry].Name,16,'');
  end;

  DirEntry:=0;
  With Dir.Entry[0] do begin
    Kind   :=$82;
    AllocSector(nTrack,nSector);
    For n:=1 to 16 do
     if n<=Length(FileName) then Name[n]:=FileName[n] else Name[n]:='';

    ATrack :=nTrack;
    ASector:=nSector;
  end;

  NewFile:=ST_OK;
 end;

Procedure TruncFile(Track,Sector:byte);
 Var
  Temp:TSector;
  UnAlloc:Longint;
 begin
  Temp.NextTrack :=Track;
  Temp.NextSector:=Sector;

  While ReadSector(Temp.NextTrack,Temp.NextSector,Temp)
    and(Temp.NextTrack>0) do begin { if there's a Sector After }

  { Unalloc the NEXT Sector }
   UnAlloc:=$100; { split in 2 lines for LongInt SHL }
   UnAlloc:=UnAlloc SHL Temp.NextSector;
   If (BAM.SAT[Temp.NextTrack] and UnAlloc)=0 then begin
    Inc(BAM.SAT[Temp.NextTrack],UnAlloc+1);
    Inc(FreeBlocks);
   end;

  end;

 end;

function SaveFile(Channel:byte):byte;
 var
  replace:boolean;
  exist:boolean;
 begin
 { invalid name }
  if (pos('*',FileName)>0)
  or (pos('?',FileName)>0) then begin
   SaveFile:=ST_TIMEOUT;
   exit;
  end;
 { replace file }
  if copy(FileName,1,3)='@0:' then begin { replace }
   Delete(FileName,1,3);
   replace:=true;
  end else
   replace:=false;
  { file exist }

  with Channels[Channel] do begin

  exist:=FindFile(FileName,Sector.NextTrack,Sector.NextSector,True);
  if exist and (not replace) then begin
   SaveFile:=ST_TIMEOUT;
   exit;
  end;

  if exist then begin
   TruncFile(Sector.NextTrack,Sector.NextSector);
   Dir.Entry[DirEntry].Blocks:=0; { clear file }
   SaveFile:=ST_OK;
  end else begin
   SaveFile:=NewFile(Sector.NextTrack,Sector.NextSector);
  end;

  end;

 end;

function Load(Channel:byte; AFileName:string):byte;
 begin
  SetError(errNone,0,0);

  if (Handle=0)or(Channel>15) then
   Load:=ST_TIMEOUT
  else begin

   Channels[Channel].SectorIndex:=0;

   if Channel=15 then
    {LoadError}
   else begin

    FileName:=AFileName;
    if FileName[1]='$' then begin
     if Channel=0 then
      Load:=LoadDirectory
     else with Channels[Channel] do begin
      Sector.NextTrack:=18;
      Sector.NextSector:=0;
      Load:=ReadNextSector(Channel);
     end;
    end else
     if FileName='#' then Load:=ST_OK else Load:=LoadFile(Channel);

   end;
  end;
 end;

function ReadByte(Channel:byte):byte;
 begin

{  statusbar('READ BYTE '+chr(ord('0')+Channel)); }

  if (Channel<15)and(Handle=0) then begin
  {$IFNDEF DEBUG}
   StatusBar('No drive');
  {$ENDIF}
   ReadByte:=$40;{T_NOTPRESENT}
   exit;
  end;

  with Channels[Channel] do begin

  if (SectorIndex>=SectorSize) then begin
  {$IFNDEF DEBUG}
   StatusBar('Reading error ');
  {$ENDIF}
   ReadByte:=ST_READ_TIMEOUT;
   exit;
  end;

  if SectorSize=256 then begin
   A:=Sector.NextTrack;
   dec(SectorSize);
  end else begin
   if SectorSize=255 then begin
    A:=Sector.NextSector;
    dec(SectorSize);
   end else begin
    A:=Sector.Data[SectorIndex];
    inc(SectorIndex);
   end;
  end;

  ReadByte:=ST_OK;

  if SectorIndex>=SectorSize then begin

   Case FillSector of
    fsNone : ReadByte:=ST_FILENOTFOUND;
    fsRead : ReadNextSector(Channel);
    fsDir  : NextDirectory;
    fsDone : begin ReadByte:=ST_EOF; SetError(errNone,0,0); end;
      else   ReadByte:=ST_READ_TIMEOUT;
   end;

  end;

  end;

 end;

function inttostr(c:byte):string;
 var
  s:string;
 begin
  str(c,s);
  inttostr:=s;
 end;

Procedure ExecCommand;
{
 + NEW:DISKNAME,DISKID               : format a new disk (create a new .d64)
 + RENAME:NEWFILE=OLDFILE            : rename
 + SCRATCH:FILENAME                  : delete
 - COPY:NEWFILE=OLDFILE              : copy file
 - COPY:NEWFILE=OLD1,OLD2,OLD3,OLD4  : concact up to 4 files
 - INITIALIZE                        : reset 1541
 - VALIDATE                          : flush buffers

 + U1:channel; drive; track; block
}
 var
  p :byte;
  ID:word;
  oldname:string; c,t,s:byte;
  a :longint;

  procedure unknow;
   begin
   {$IFNDEF DEBUG}
{    StatusBar(Command); }
    GetKey;
   {$ENDIF}
   end;

 begin
  {$IFNDEF DEBUG}
   StatusBar(Command);
  {$ENDIF}
  case Command[1] of
  { NEW }
   'N' : begin
          if Command[2]=':' then Delete(Command,1,2) else
          if Copy(Command,1,4)='NEW:' then Delete(Command,1,4) else begin unknow; exit end;
          p:=pos(',',Command);
          if p=0 then
           ID:=ord('0')+256*ord('0')
          else begin
           move(command[p+1],ID,2);
           Command[0]:=chr(p-1);
          end;
          CloseDevice;
          D64File:=copy(Command,1,8);
          for p:=1 to length(D64File) do
           if not (D64File[p] in ['0'..'9','A'..'Z']) then D64File[p]:='_';
          D64File:=D64File+'.D64';
          CreateDevice(Command,ID);
         end;
  { RENAME }
   'R' : begin
          if not OpenDevice then begin unknow; exit; end;
          if Command[2]=':' then Delete(Command,1,2) else
          if Copy(Command,1,7)='RENAME:' then Delete(Command,1,7) else begin unknow; exit; end;
          p:=pos('=',Command); if p=0 then begin unknow; exit; end;
          oldname:=copy(command,p+1,length(command));
          if not findfile(oldname,t,s,true) then begin unknow; exit; end;
          command[0]:=chr(p-1); { newname }
          if pos('*',Command)>0 then begin unknow; exit; end;
          if pos('?',Command)>0 then begin unknow; exit; end;
          with Dir.Entry[DirEntry] do begin
           For p:=1 to 16 do
            if p<=Length(command) then Name[p]:=command[p] else Name[p]:='';
          end;
          WriteSector(DirTrack,DirSector,Dir);
         end;
   { SCRATCH }
   'S' : begin
          if not OpenDevice then begin unknow; exit; end;
          if Command[2]=':' then Delete(Command,1,2) else
          if Copy(Command,1,8)='SCRATCH:' then Delete(Command,1,8) else begin unknow; exit; end;
          while findfile(command,t,s,false) do begin { you can use jocker }
           TruncFile(t,s);
           a:=$100; a:=a shl s;
           if (BAM.SAT[t] and a)=0 then begin
            inc(BAM.SAT[t],a+1);
            inc(FreeBlocks);
           end;
           FillChar(Dir.Entry[DirEntry],SizeOf(TDirEntry),0); { clear file }
           WriteSector(DirTrack,DirSector,Dir);
           WriteSector(18,0,BAM);
          end;
         end;

   { USER }
   'U' : case Command[2] of

          { U1:channel; drive; track; block }
          { U1 5 0 18 00 }
          '1','A': begin
           if not OpenDevice then begin unknow; exit; end;
           c:=ord(Command[4])-ord('0');
           p:=8;

           t:=0;
           while Command[p] in ['0'..'9'] do begin
            t:=10*t+ord(Command[p])-ord('0');
            inc(p);
           end;
           inc(p);
           s:=0;
           while (p<length(command))and(Command[p] in ['0'..'9']) do begin
            s:=10*s+ord(Command[p])-ord('0');
            inc(p);
           end;

           with Channels[c] do begin
            ReadSector(t,s,Sector);
            SectorSize:=256; { ??? }
            SectorIndex:=0;
            FillSector:=fsDone;
           end;

          end;

          else unknow;

         end;

   else unknow;
  end;
  Command:='';
 end;

function WriteByte(Channel,Value:byte; EndOfFile:boolean):byte;
 Var
  T,S:byte;
 begin
  WriteByte:=ST_OK;

  if Channel=15 then begin
   if not EndOfFile then Command:=Command+Chr(Value) else ExecCommand;
   exit;
  end;

  with Channels[Channel] do begin

  if (Handle=0)or(SectorIndex>253) then begin
   WriteByte:=ST_TIMEOUT;
  {$IFNDEF DEBUG}
   StatusBar('Write Error !');
   GetKey;
  {$ENDIF}
   SetError(errWrite,Sector.NextTrack,Sector.NextSector);
   exit;
  end;

  Sector.Data[SectorIndex]:=Value;
  Inc(SectorIndex);

  SetError(errNone,0,0);

  if (SectorIndex>253)Or(EndOfFile) then begin

   T:=Sector.NextTrack;   { current Sector }
   S:=Sector.NextSector;

   if EndOfFile then begin
    Sector.NextTrack :=0;   { last sector }
    Sector.NextSector:=SectorIndex+1;
   end else begin
    AllocSector(Sector.NextTrack,Sector.NextSector); { next sector }
   end;

   WriteSector(T,S,Sector);
   Inc(Dir.Entry[DirEntry].Blocks);

   SectorIndex:=0;

   If EndOfFile then begin
   { Flush }
    WriteSector(18,0,BAM); { for AllocSector }
    WriteSector(DirTrack,DirSector,Dir); { for Dir.Entry[DirEntry].Blocks }
   end;

  end;

  end;
 end;


function Save(Channel:byte; AFileName:string):byte;
 begin
  if (Handle=0)or(Channel>11) then
   Save:=ST_TIMEOUT
  else begin
   FileName:=AFileName;
   Channels[Channel].SectorIndex:=0;
   if FileName='$' then Save:=ST_TIMEOUT else Save:=SaveFile(Channel);
  end;
 end;


begin
 Handle :=0;
 D64File:='';

 fillchar(Channels,SizeOf(Channels),0);

 SetError(ErrVersion,0,0);
(*
 FillSector:=fsNone;
 SectorIndex:=0;
*)
end.