Unit RESID; { SID Emulation, from reSID }

interface

uses dos,data,dialogs,keyboard;

Const
 SampleRate = 10000; { MHz }
 SampleTime = 256 - (1000000 div SampleRate);
 HiSpeed:boolean=(1000000 div (256-SampleTime))>22222;

 SndSize    = 64;     { sampling buffer size }
 BufSize    = 2*SndSize; { DMA ring-buffer size }

 VolConst   = 15;

 CycleRate  = 9867 div (SampleRate div SndSize); { PAL }

Var
 ADR,IRQ,DMA:word; { BLASTER=Axxx Ix Dx }

Procedure Init;
Procedure Done;

Procedure PokeSID; { called by POKE.ASM  }

const
 IRQs:word=0;

implementation

const
 ATTACK =0;
 DECAY  =1;
 SUSTAIN=1;
 RELEASE=2;

(*
 ADSR:array[0..15] of word=( { duration in ms }
  2,8,16,24,38,56,68,80,100,250,500,800,1000,3000,5000,8000
 );
*)
 TIMES:array[0..15] of word=( { ADSR*1000/256 }
  9,32,63,95,149,220,267,313,392,977,1954,3126,3907,11720,19532,31251
 );

 EXPTIME:array[0..$FF] of byte=(
  { $00: }   1, 30, 30, 30, 30, 30, 30, 16,
  { $08: }  16, 16, 16, 16, 16, 16, 16,  8,
  { $10: }   8,  8,  8,  8,  8,  8,  8,  8,
  { $18: }   8,  8,  8,  4,  4,  4,  4,  4,
  { $20: }   4,  4,  4,  4,  4,  4,  4,  4,
  { $28: }   4,  4,  4,  4,  4,  4,  4,  4,
  { $30: }   4,  4,  4,  4,  4,  4,  4,  2,
  { $38: }   2,  2,  2,  2,  2,  2,  2,  2,
  { $40: }   2,  2,  2,  2,  2,  2,  2,  2,
  { $48: }   2,  2,  2,  2,  2,  2,  2,  2,
  { $50: }   2,  2,  2,  2,  2,  2,  2,  2,
  { $58: }   2,  2,  2,  2,  2,  2,  1,  1,
  { $60: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $68: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $70: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $78: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $80: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $88: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $90: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $98: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $a0: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $a8: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $b0: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $b8: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $c0: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $c8: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $d0: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $d8: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $e0: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $e8: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $f0: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $f8: }   1,  1,  1,  1,  1,  1,  1,  1
 );

Type
 PSnd=^TSnd;
 TSnd=array[0..SndSize-1] of byte;

Var
 OldIRQ:pointer;

 Buf:pointer; { DMA double-buffer }
 Dbl:pointer; { double buffer adr }
 Snd:PSnd;    { sound buffer }
 Double:boolean;

{---DMA---}
Const
 DMAADRS:array[0..7] of byte=($00,$02,$04,$06,$C0,$C4,$C8,$CC);
 DMASIZE:array[0..7] of byte=($01,$03,$05,$07,$C2,$C6,$CA,$CE);
 DMAPAGE:array[0..7] of byte=($87,$83,$81,$82,$00,$8B,$89,$8A);

Var
 MSK:byte; { $0A or $D4   }
 MDE:byte; { $0B or $D6   }
 FLP:byte; { $0C or $D8   }
 DAD:byte; { DMAADRS[DMA] }
 DSZ:byte; { DMASIZE[DMA] }
 DPG:byte; { DMAPAGE[DMA] }

Procedure GetDMAMem(var p:pointer; Size:word);
 var
  dmapage:byte;
  dmasize:longint;
  p2     :pointer;
  size2  :word;
 begin
  GetMem(p,Size); { allocate size bytes }
  if p=nil then halt;

  dmapage:=(seg(p^)+ofs(p^) shr 4) shr 12;
  dmasize:=(seg(p^) shl 4+ofs(p^)) and $FFFF;

  if dmasize+size>$FFFF then begin { page boundary }
   freemem(p,size);
   size2:=$FFFF-dmasize;
   getmem(p2,size2);
   if p2=nil then halt;
   getmem(p,size);
   if p=nil then halt;
   freemem(p2,size2);
  end;
 end;

{---DMA (8bits) ---}

Procedure SetupDMA(DMA,Mode:byte; Buff:pointer; Size:word);
 Var
  Len :record
   case boolean of
   true :(long:longint);
   false:(lo,hi:word);
  end;
  Page:word;
  Ofst:word;
 begin
  if DMA<4 then begin
   MSK:=$0A;
   MDE:=$0B;
   FLP:=$0C;
  end else begin
   MSK:=$D4;
   MDE:=$D6;
   FLP:=$D8;
  end;
  DAD:=DMAADRS[DMA];
  DSZ:=DMASIZE[DMA];
  DPG:=DMAPAGE[DMA];

  len.long:=longint(seg(buff^)) shl 4+ofs(buff^);
  page:=len.hi;
  ofst:=len.lo;

  write(#16' DMA Setup :');
  write(' DMA=',DMA);
  write(' MSK=',HexaB(MSK));
  write(' FLP=',HexaB(FLP));
  write(' ADR=',HexaB(DAD));
  write(' PGE=',HexaB(DPG));
  write(' SIZ=',HexaB(DSZ));
  writeln(' MOD=',HexaB(MDE));

  writeln(#254' PGE=',HexaW(Page),' OFS=',HexaW(ofst),' SIZE=',HexaW(SIZE));

  DMA:=DMA and 3;

  port[MSK]:=DMA or 4; { disable }

  port[DAD]:=lo(ofst);
  port[DAD]:=hi(ofst);
  port[FLP]:=0;        { reset }
  port[DSZ]:=lo(size);
  port[DSZ]:=hi(size);
  port[DPG]:=lo(page);
  if page>255 then port[DPG+$400]:=hi(page);

  port[MDE]:=DMA or Mode;
  port[MSK]:=DMA;      { enable }
 end;

{---IRQ---}
Procedure EnableIRQ(IRQ:byte);
 begin
  Port[$21]:=(Port[$21] and (255 - ( 1 shl IRQ)));
 end;

Procedure DisableIRQ(IRQ:byte);
 begin
  port[$21]:=port[$21] or (1 shl IRQ);
 end;

{---SoundBlaster---}

Function SbSettings:boolean;
 var
  Blaster:string;
  p:byte;
 begin
  Write(#16' Init Sound Card - ');

  Blaster:=GetEnv('BLASTER'); { BLASTER=A220 I5 D1 H5 }
  SbSettings:=False;
  p:=pos('A',Blaster); if p=0 then exit;
  ADR:=$200+16*(ord(Blaster[p+2])-ord('0')); { A2x0 }
  p:=pos('I',Blaster); if p=0 then exit;
  IRQ:=ord(Blaster[p+1])-ord('0');            { I5 }
  p:=pos('D',Blaster); if p=0 then exit;
  DMA:=ord(Blaster[p+1])-ord('0');            { D1 }

  SbSettings:=True;

  Blaster:='BLASTER='+HexaW(ADR);
  Blaster[9]:='A';

  WriteLn(Blaster,' I',IRQ,' D',DMA);
 end;

Function SBReset:boolean;
 var
  i:integer;
  x:byte;
 begin
  Port[ADR+6]:=1; for i:=0 to 3 do x:=Port[ADR+6];
  Port[ADR+6]:=0;
  SBReset:=True;
  for i:=0 to 100 do
   if (Port[ADR+$E] and $80)>0 then { Status }
    if Port[ADR+$A]=$AA then exit;  { Sign   }
  SBReset:=False;
 end;

function SBReadDSP:integer;
 begin
  repeat until (Port[ADR+$E] and $80)>0; { Status }
  SBReadDSP:=Port[ADR+$A]; { Read }
 end;

Procedure SBWriteDSP(Value:byte);
 begin
  repeat until (Port[ADR+$C] and $80)=0; { Write enable }
  Port[ADR+$C]:=Value; { Write }
 end;

{ ------------- SID Emulation --------------- }

Type
 TVoice=record
 { SyncSource }
  Source :word;
 { Envelope }
  Play   :boolean;
  A,D,S,R:word; { ATTACK, DECAY, SUSTAIN, RELEASE level }
  Gate   :byte;
  Action :word; { A/D/S/R }
  Timer  :word; { Clock ticks }
  Alarm  :word;
  ExpTm  :word;
  Volume :word; { Volume Level }
  Level  :word; { S+16*S }
 { Wave }
  Freq   :word;
  Pulse  :word;
  Wave   :word; { WAVE type (can be a mixed type) }

  Ring   :boolean;
  Sync   :boolean;
  Test   :boolean;
  Rise   :boolean;

  Output :longint;

  Noise  :longint;
  Accum  :longint;
 end;

var
 MaxVol  :word;
 Voices:array[0..2] of TVoice;
 Keys  :byte;
 Mix:array[0..SndSize-1] of longint;

 i:integer;

 OldAccum:longint;
 bit0:longint;

Procedure EnvelopeClock(Voice:word);
 begin
  with Voices[Voice] do begin

   Inc(Timer);

   if (Timer and $7FFF)<>Alarm then exit;

   if (Timer and $8000)>0 then begin
    Timer:=1; { second loop }
    exit;
   end;

   Timer:=0; { reset Timer }
   Inc(ExpTM);

   if (Action=ATTACK)or(ExpTM=EXPTIME[Volume]) then begin
    ExpTM:=0;
    if not Play then exit;

    Case Action of
     ATTACK : begin
               Volume:=Succ(Volume) and $FF;
               if Volume=$FF then begin
                Action:=DECAY;
                Alarm :=TIMES[D];
               end;
              end;
     DECAY  : begin
               if Volume<>Level then dec(Volume);
              end;
     RELEASE: begin
               Volume:=Pred(Volume) and $FF;
              end;
    end;
   end;

   if Volume=0 then Play:=False;

  end;
 end;

Var
 Index2:word;
procedure WaveClock(Voice:word);
 begin
  with Voices[Voice] do begin

   if Test then exit;

   OldAccum:=Accum;
   Accum:=(Accum+Freq) and $FFFFFF;
   Rise:=((OldAccum and $80000)=0)and((Accum and $80000)>0);

   if Rise then begin
    bit0 :=((Noise shr 22) xor (Noise shr 17)) and 1;
    Noise:=((Noise shl 1) and $7FFFFF) or bit0;
   end;

  end;
 end;

Procedure SyncWave(Voice:word);
 begin
  with Voices[Voice] do begin
   if Sync and (Voices[Source].Rise) then Accum:=0;
  end;
 end;

Procedure WaveOutput(Voice:word);
 var
  msb:boolean;
 begin
  With Voices[Voice] do begin

  { None or Noise combinations }
   if (Wave=0)or(Wave>8) then begin
    Output:=0;
    exit;
   end;

  { Noise only }
   if Wave=8 then begin
    Output:=((Noise and $400000) shr 11) or
            ((Noise and $100000) shr 10) or
            ((Noise and $010000) shr  7) or
            ((Noise and $002000) shr  5) or
            ((Noise and $000800) shr  4) or
            ((Noise and $000080) shr  1) or
            ((Noise and $000010) shl  1) or
            ((Noise and $000004) shl  2);
    exit;
   end;

  { Pulse }
   if (Wave and 4)>0 then begin
    if Test and ((Accum shr 12)<Pulse) then begin
     Output:=0; { AND anything still 0 ! }
     exit;
    end;
   end;

  { Output mask }
   Output:=$FFF;

  { SawTooth }
   if (Wave and 2)>0 then Output:=Output and (Accum shr 12);

  { Triangle }
   if (Wave and 1)>0 then begin

    if Ring then
     msb:=((Accum xor Voices[Source].Accum) and $800000)>0

    else
     msb:=(Accum and $800000)>0;

    if msb then
     Output:=Output and ((not Accum) shr 11)
    else
     Output:=Output and (Accum shr 11);

   end;

  end;
 end;

Var
 CPU:byte;
 VoiceCount:byte;
 var
  t:longint;

procedure SbIRQ8; interrupt;
 begin
  CPU:=Test8086;
  Test8086:=0;   { BP7 bug on 386+ Longint math (use 32bits regs not saved) }

  i:=Port[ADR+$E]; { Interruption Ok }
  Port[$20]:=$20;      { End of IRQ }

  if HiSpeed then SBWriteDSP($48) else SBWriteDSP($14);
  SBWriteDSP(lo(SndSize-1));
  SBWriteDSP(hi(SndSize-1));
  if HiSpeed then SBWriteDSP($91);

  inc(IRQs);

  if double then begin
   Snd:=Dbl;
   double:=false;
  end else begin
   Snd:=Buf;
   double:=true;
  end;

  for i:=0 to SndSize-1 do begin

   EnvelopeClock(0);
   EnvelopeClock(1);
   EnvelopeClock(2);

   WaveClock(0);
   WaveClock(1);
   WaveClock(2);

   SyncWave(0);
   SyncWave(1);
   SyncWave(2);

   WaveOutput(0);
   WaveOutput(1);
   WaveOutput(2);

   VoiceCount:=0;
   t:=0;
   with Voices[0] do if Play then begin inc(VoiceCount); Inc(t,(Output-2048)*Volume); end;
   with Voices[1] do if Play then begin inc(VoiceCount); Inc(t,(Output-2048)*Volume); end;
   with Voices[2] do if Play then begin inc(VoiceCount); Inc(t,(Output-2048)*Volume); end;

   if VoiceCount=0 then
    t:=0
   else begin
    t:=(t div VoiceCount) div $FFF;
    if t<-128 then t:=-128;
    t:=(t+128)*MaxVol div VolConst;
    if t>255 then t:=255;
   end;
   Snd^[i]:=t;
  end;

  Test8086:=CPU;

 end;

{---VIC---}

Procedure Init;
 var
  i:integer;
  t:longint;
  offset:longint;
  page  :word;
  size  :word;
 begin
  if not SbSettings then halt;
  if not SbReset then halt;

  Keys:=0;
  FillChar(Voices,SizeOf(Voices),0);
  Voices[0].Noise:=$7FFFF8;
  Voices[1].Noise:=$7FFFF8;
  Voices[2].Noise:=$7FFFF8;
  Voices[0].Source:=2;
  Voices[1].Source:=0;
  Voices[2].Source:=1;

  GetDMAMem(Buf,BufSize);  { DMA double-buffer }
  FillChar(Buf^,BufSize,0);

  Dbl:=ptr(Seg(Buf^),Ofs(Buf^)+SndSize);
  Double:=False;

 { IRQ }
  GetIntVec(IRQ+8,OldIRQ);

 { DMA }
  SetIntVec(IRQ+8,@SbIRQ8);
  SetupDMA(DMA,$58,Buf,BufSize-1);

  Readln;

 { DSP - Sets sample rate through internal I/O transfer timer. }
  SBWriteDSP($40);
  SBWriteDSP(SampleTime);

 { Enables speaker output. }
  SBWriteDSP($D1);

 { DSP - Set DMA transfer size }
  if HiSpeed then SBWriteDSP($48) else SBWriteDSP($14);
  SBWriteDSP(lo(SndSize-1));
  SBWriteDSP(hi(SndSize-1));
  if HiSpeed then SBWriteDSP($91);

  EnableIRQ(IRQ);
 end;

Procedure Done;
 begin
  DisableIRQ(IRQ);
  SetIntVec(IRQ+8,OldIRQ);
  SBWriteDSP($D3);
  SBReset;
  FreeMem(Buf,BufSize);
 end;

Function PitchToRatio(Pitch:word):word;
 begin
  PitchToRatio:=longint(Pitch) shl 14 div SampleRate;
 end;

Procedure VoiceControl(Voice:word);
 Var
  NewGate:byte;
 begin
  with Voices[Voice] do begin

  { Envelope }

   NewGate:=SIDPoke and 1;

   if (Gate=0)and(NewGate=1) then begin
    Action:=ATTACK;
    Alarm :=TIMES[A];
    Play  :=True;
   end;

   if (Gate=1)and(NewGate=0) then begin
    Action:=RELEASE;
    Alarm :=TIMES[R];
   end;

   Gate:=NewGate;

  { Wave }

   Wave:=(SIDPoke shr 4) and $F;

   Ring:=(SIDPoke and 4)>0;
   Sync:=(SIDPoke and 2)>0;
   Test:=(SIDPoke and 8)>0;

   if Test then begin
    Accum:=0;
    Noise:=0;
   end else begin
    Noise:=$7FFFF8
   end;

  end;
 end;

Procedure PokeSID;
 begin
  case SIDAddr of
{ VOICE 0 }
   0,1: Voices[0].Freq :=MemW[CIOSeg:$D400];
   2,3: Voices[0].Pulse:=MemW[CIOSeg:$D402] and $FFF;

   4  : VoiceControl(0);

   5  : With Voices[0] do begin
         A:=SIDPoke shr 4;
         D:=SIDPoke and 15;
         if Action=ATTACK then Alarm:=TIMES[A] else
         if Action=DECAY  then Alarm:=TIMES[D];
        end;

   6  : With Voices[0] do begin
         S:=SIDPoke shr 4;
         R:=SIDPoke and 15;
         Level:=S+16*S;
         if Action=RELEASE then Alarm:=TIMES[R];
        end;

{ VOICE 1 }
   7,8 : Voices[1].Freq :=MemW[CIOSeg:$D400+7];
   9,10: Voices[1].Pulse:=MemW[CIOSeg:$D400+9] and $FFF;

   11  : VoiceControl(1);

   12  : with Voices[1] do begin
          A:=SIDPoke shr 4;
          D:=SIDPoke and 15;
          if Action=ATTACK then Alarm:=TIMES[A] else
          if Action=DECAY  then Alarm:=TIMES[D];
         end;

   13  : with Voices[1] do begin
          S:=SIDPoke shr 4;
          R:=SIDPoke and 15;
          Level:=S+16*S;
          if Action=RELEASE then Alarm:=TIMES[R];
         end;

{ VOICE 2 }
   14,15: Voices[2].Freq :=MemW[CIOSeg:$D400+14];
   16,17: Voices[2].Pulse:=MemW[CIOSeg:$D400+16] and $FFF;

   18   : VoiceControl(2);

   19   : With Voices[2] do begin
           A:=SIDPoke shr 4;
           D:=SIDPoke and 15;
           if Action=ATTACK then Alarm:=TIMES[A] else
           if Action=DECAY  then Alarm:=TIMES[D];
          end;

   20   : With Voices[2] do begin
           S:=SIDPoke shr 4;
           R:=SIDPoke and 15;
           Level:=S+16*S;
           if Action=RELEASE then Alarm:=TIMES[R];
          end;

{ GENERAL }
  {21,22: FC (word)}
  {32   :RES_FILT(byte)}
   24: MaxVol:=(SIDPoke and 15);
  end;
 end;
end.