Unit RESID9; { SID Emulation from reSID by Dag Lem <resid@nimrod.no>  }

interface

uses dos,data,dialogs,keyboard;

Const
 SampleRate = 22000; { MHz }
 SampleTime = 256 - (1000000 div SampleRate);
 HiSpeed:boolean=(1000000 div (256-SampleTime))>22222;

 SndSize    = 64;      { sampling buffer size }
 BufSize    = 2*SndSize; { DMA ring-buffer size }

 VolConst   = 15;

 CycleRate  = 9867 div (SampleRate div SndSize); { PAL }

Var
 ADR,IRQ,DMA:word; { BLASTER=Axxx Ix Dx }

Procedure Init;
Procedure Done;

Procedure PokeSID; { called by POKE.ASM (set SIDAddr & SIDPoke }
Procedure SIDLine;

const
 IRQs:word=0;

implementation

const
 ATTACK =0;
 DECAY  =1;
 SUSTAIN=1; { nothing special, like DECAY }
 RELEASE=2;
(*
 ADSR:array[0..15] of word=( { duration in ms }
  2,8,16,24,38,56,68,80,100,250,500,800,1000,3000,5000,8000
 );
*)
 TIMES:array[0..15] of word=( { ADSR*1000/256 }
  9,32,63,95,149,220,267,313,392,977,1954,3126,3907,11720,19532,31251
 );

 EXPTIMES:array[0..$FF] of byte=(
  { $00: }   1, 30, 30, 30, 30, 30, 30, 16,
  { $08: }  16, 16, 16, 16, 16, 16, 16,  8,
  { $10: }   8,  8,  8,  8,  8,  8,  8,  8,
  { $18: }   8,  8,  8,  4,  4,  4,  4,  4,
  { $20: }   4,  4,  4,  4,  4,  4,  4,  4,
  { $28: }   4,  4,  4,  4,  4,  4,  4,  4,
  { $30: }   4,  4,  4,  4,  4,  4,  4,  2,
  { $38: }   2,  2,  2,  2,  2,  2,  2,  2,
  { $40: }   2,  2,  2,  2,  2,  2,  2,  2,
  { $48: }   2,  2,  2,  2,  2,  2,  2,  2,
  { $50: }   2,  2,  2,  2,  2,  2,  2,  2,
  { $58: }   2,  2,  2,  2,  2,  2,  1,  1,
  { $60: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $68: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $70: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $78: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $80: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $88: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $90: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $98: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $a0: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $a8: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $b0: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $b8: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $c0: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $c8: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $d0: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $d8: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $e0: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $e8: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $f0: }   1,  1,  1,  1,  1,  1,  1,  1,
  { $f8: }   1,  1,  1,  1,  1,  1,  1,  1
 );

Type
 PSnd=^TSnd;
 TSnd=array[0..SndSize-1] of byte;

Var
 OldIRQ:pointer;

 Buf:pointer; { DMA double-buffer }
 Dbl:pointer; { double buffer adr }
 Snd:PSnd;    { sound buffer }
 Double:boolean;

{---DMA---}
Const
 DMAADRS:array[0..7] of byte=($00,$02,$04,$06,$C0,$C4,$C8,$CC);
 DMASIZE:array[0..7] of byte=($01,$03,$05,$07,$C2,$C6,$CA,$CE);
 DMAPAGE:array[0..7] of byte=($87,$83,$81,$82,$00,$8B,$89,$8A);

Var
 MSK:byte; { $0A or $D4   }
 MDE:byte; { $0B or $D6   }
 FLP:byte; { $0C or $D8   }
 DAD:byte; { DMAADRS[DMA] }
 DSZ:byte; { DMASIZE[DMA] }
 DPG:byte; { DMAPAGE[DMA] }

Procedure GetDMAMem(var p:pointer; Size:word);
 var
  dmapage:byte;
  dmasize:longint;
  p2     :pointer;
  size2  :word;
 begin
  GetMem(p,Size); { allocate size bytes }
  if p=nil then halt;

  dmapage:=(seg(p^)+ofs(p^) shr 4) shr 12;
  dmasize:=(seg(p^) shl 4+ofs(p^)) and $FFFF;

  if dmasize+size>$FFFF then begin { page boundary }
   freemem(p,size);
   size2:=$FFFF-dmasize;
   getmem(p2,size2);
   if p2=nil then halt;
   getmem(p,size);
   if p=nil then halt;
   freemem(p2,size2);
  end;
 end;

{---DMA (8bits) ---}

Procedure SetupDMA(DMA,Mode:byte; Buff:pointer; Size:word);
 Var
  Len :record
   case boolean of
   true :(long:longint);
   false:(lo,hi:word);
  end;
  Page:word;
  Ofst:word;
 begin
  if DMA<4 then begin
   MSK:=$0A;
   MDE:=$0B;
   FLP:=$0C;
  end else begin
   MSK:=$D4;
   MDE:=$D6;
   FLP:=$D8;
  end;
  DAD:=DMAADRS[DMA];
  DSZ:=DMASIZE[DMA];
  DPG:=DMAPAGE[DMA];

  len.long:=longint(seg(buff^)) shl 4+ofs(buff^);
  page:=len.hi;
  ofst:=len.lo;

  write(#16' DMA Setup :');
  write(' DMA=',DMA);
  write(' MSK=',HexaB(MSK));
  write(' FLP=',HexaB(FLP));
  write(' ADR=',HexaB(DAD));
  write(' PGE=',HexaB(DPG));
  write(' SIZ=',HexaB(DSZ));
  writeln(' MOD=',HexaB(MDE));

  writeln(#254' PGE=',HexaW(Page),' OFS=',HexaW(ofst),' SIZE=',HexaW(SIZE));

  WriteLn('---press ENTER---'); Readln;

  DMA:=DMA and 3;

  port[MSK]:=DMA or 4; { disable }

  port[DAD]:=lo(ofst);
  port[DAD]:=hi(ofst);
  port[FLP]:=0;        { reset }
  port[DSZ]:=lo(size);
  port[DSZ]:=hi(size);
  port[DPG]:=lo(page);
  if page>255 then port[DPG+$400]:=hi(page);

  port[MDE]:=DMA or Mode;
  port[MSK]:=DMA;      { enable }
 end;

{---IRQ---}
Procedure EnableIRQ(IRQ:byte);
 begin
  Port[$21]:=(Port[$21] and (255 - ( 1 shl IRQ)));
 end;

Procedure DisableIRQ(IRQ:byte);
 begin
  port[$21]:=port[$21] or (1 shl IRQ);
 end;

{---SoundBlaster---}

Function SbSettings:boolean;
 var
  Blaster:string;
  p:byte;
 begin
  Write(#16' Init Sound Card - ');

  Blaster:=GetEnv('BLASTER'); { BLASTER=A220 I5 D1 H5 }
  SbSettings:=False;
  p:=pos('A',Blaster); if p=0 then exit;
  ADR:=$200+16*(ord(Blaster[p+2])-ord('0')); { A2x0 }
  p:=pos('I',Blaster); if p=0 then exit;
  IRQ:=ord(Blaster[p+1])-ord('0');            { I5 }
  p:=pos('D',Blaster); if p=0 then exit;
  DMA:=ord(Blaster[p+1])-ord('0');            { D1 }

  SbSettings:=True;

  Blaster:='BLASTER='+HexaW(ADR);
  Blaster[9]:='A';

  WriteLn(Blaster,' I',IRQ,' D',DMA);
 end;

Function SBReset:boolean;
 var
  i:integer;
  x:byte;
 begin
  Port[ADR+6]:=1; for i:=0 to 3 do x:=Port[ADR+6];
  Port[ADR+6]:=0;
  SBReset:=True;
  for i:=0 to 100 do
   if (Port[ADR+$E] and $80)>0 then { Status }
    if Port[ADR+$A]=$AA then exit;  { Sign   }
  SBReset:=False;
 end;

function SBReadDSP:integer;
 begin
  repeat until (Port[ADR+$E] and $80)>0; { Status }
  SBReadDSP:=Port[ADR+$A]; { Read }
 end;

Procedure SBWriteDSP(Value:byte);
 begin
  repeat until (Port[ADR+$C] and $80)=0; { Write enable }
  Port[ADR+$C]:=Value; { Write }
 end;

{ ------- SID Emulation -------- }

Type
 TVoice=record
  Wave   :word;    { WAVE type (can be a mixed type) }
  Index  :word;    { Index in this WAVE }
  Volume :word;    { Volume Level }
  A,D,S,R:word;    { ATTACK, DECAY, SUSTAIN, RELEASE level }
  Freq   :word;
  Pulse  :word;
  Action :word;    { A/D/S/R }
  Timer  :word;
  ExpTime:word;
  Alarm  :word;
  Level  :word;
  Noise  :word;
  Source :word;
  Output :integer;

  Gate   :boolean;
  Test   :boolean;
  Ring   :boolean;
  Sync   :boolean;
  Rise   :boolean;
  Filter :boolean;
  Off    :boolean; { only voice 3 }
  Play   :boolean;
 end;

var
 MaxVol:word;
 Voices:array[0..2] of TVoice;

 FilterFreq:word;
 Filter:byte;

 i:integer;

 msb:boolean;
 Index2:word;

Var
 Mix:array[0..SndSize-1] of longint;
 Count:array[0..SndSize-1] of byte;
 OutMix    :longint;
 VoiceCount:word;

procedure MixVoice(Voice:word);
 begin
  with Voices[Voice] do begin         { with this voice }

  for i:=0 to 63 do begin

   {------------------- Envelope Clock -------------------}
    Inc(Timer);

{    if (Timer and $8000)>0 then Timer:=1; { there ? }

    if (Timer and $7FFF)=Alarm then begin

     if (Timer and $8000)>0 then { or there ? }

      Timer:=1

     else begin

      Timer:=0;
      Inc(ExpTime);

      if (Action=ATTACK)or(ExpTime=EXPTIMES[Volume]) then begin

       ExpTime:=0;

       if Play then begin

        Case Action of
         ATTACK  : begin
                    Volume:=Succ(Volume) and $FF;
                    if Volume=$FF then begin
                     Action:=DECAY;
                     Alarm :=TIMES[D];
                    end;
                   end;

         DECAY   : if Volume<>Level then Volume:=Pred(Volume) and $FF;

         RELEASE : Volume:=Pred(Volume) and $FF;

        end; { Case Action of }

        if Volume=0 then Play:=False;

       end; { if Play }

      end; { ATTACK or ExpTime }

     end; { Timer and $8000 }

    end; { Timer = Alarm }

   {-------------------- Wave clock ----------------------}

    if Not Test then begin
     Index2:=Index;
     Inc(Index,(Freq shr 4)+1);
     Rise:=((Index2 and $8000)=0) and ((Index and $8000)>0);
     if Rise then
      Noise:=(Noise shl 1)or(((Noise shr 15)xor(Noise shr 11)) and 1);
    end; { not test }

   {------------------- Wave Sync ------------------------}
    if Sync and (Voices[Source].Rise) then Index:=0;

   end; { for i:=0 to SndSize-1 }

  end; { with voices[voice] }
 end;

Var
 vMix:longint;
 vCount:word;

Procedure VoiceOutput(Voice:word);
 begin
  with Voices[Voice] do begin
   {-------------------- Output --------------------------}
    if (Play)and(Wave>0)and(Volume>0)and(Not Off) then begin

     if Wave=8 then { Noise, get a random value }

      Output:=(((Noise and $4000) shr 7) or
               ((Noise and $1000) shr 6) or
               ((Noise and $0100) shl 3) or
               ((Noise and $0020) shr 1) or
               ((Noise and $0008))
              )and 255

{
    Output:=((Noise and $4000-00) shr 11) or
            ((Noise and $1000-00) shr 10) or
            ((Noise and $0100-00) shr  7) or
            ((Noise and $0020-00) shr  5) or
            ((Noise and $0008-00) shr  4) or
            ((Noise and $0000-80) shr  1) or
            ((Noise and $0000-10) shl  1) or
            ((Noise and $0000-04) shl  2);
}
     else begin { Not Noise }

      if ((Wave and 4)>0)and(Not Test)and(Index<Pulse) then begin { SQR }

       Output:=0;

      end else begin

       Output:=$FF;

       if (Wave and 2)>0 then Output:=Output and Hi(Index); { SAW }

       if (Wave and 1)>0 then begin { TRI }

        if Ring then
         msb:=((Index xor Voices[Source].Index) and $8000)>0
        else
         msb:=(Index xor $8000)>0;

        if msb then
         Output:=Output and (not Hi(Index))
        else
         Output:=Output and Hi(index);

       end; { no TRI }

      end; { not SQR0 }

     end; { if Wave=8 }

     Inc(vMix,longint(Output-128)*Volume); { integer:(0..255)-128*(0..255) }
     Inc(vCount);

    end; { if Play }

  end; { with voices[voice] }

 end;

Var
 CPU:byte;
Const
 vMax=3*255*255;
var
 t:integer;

procedure SbIRQ8; interrupt;
 begin
  CPU:=Test8086;
  Test8086:=0;   { BP7 bug on Longint math (use 32bits regs not saved) }

  i:=Port[ADR+$E]; { Interruption Ok }
  Port[$20]:=$20;  { End of IRQ }

  if HiSpeed then SBWriteDSP($48) else SBWriteDSP($14);
  SBWriteDSP(lo(SndSize-1));
  SBWriteDSP(hi(SndSize-1));
  if HiSpeed then SBWriteDSP($91);

  inc(IRQs);

{ Inc(Mem[CIOSeg:53280]); }

  if double then begin
   Snd:=Dbl;
   double:=false;
  end else begin
   Snd:=Buf;
   double:=true;
  end;

{  FillChar(Mix,SizeOf(Mix),0); }
  asm
   mov ax,SEG @DATA
   mov es,ax
   xor ax,ax

   mov di,offset Mix
   mov cx,SndSize*4/2
   rep stosw

   mov di,Offset Count
   mov cx,SndSize/2
   rep stosw
  end;

  MixVoice(0);
  MixVoice(1);
  MixVoice(2);

  for i:=0 to SndSize-1 do begin
   VoiceCount:=Count[i];
   if (VoiceCount=0)or(MaxVol=0) then
    t:=0
   else begin
    t:=(Mix[i] div VoiceCount) shr 8; { VoiceVolume : 0..255 }
    if t<-128 then t:=-128;
    t:=((t+128)*MaxVol)div VolConst;
    if t>255 then t:=255;
   end;
   Snd^[i]:=t;
  end;

  Test8086:=CPU;  { Enable 386+ }

 end;

{---SID---}
Procedure SIDLine;
 begin
  MixVoice(0);
  MixVoice(1);
  MixVoice(2);

  vMix:=0;
  vCount:=0;
  VoiceOutput(0);
  VoiceOutput(1);
  VoiceOutput(2);

  if vCount=0 then
   t:=0
  else begin
   t:=(vMix div vCount) div 255;
   if t<-128 then t:=-128;
   t:=((t+128)*MaxVol)div VolConst;
   if t>255 then t:=255;
  end;
  SBWriteDSP($10);
  SBWriteDSP(t);
 end;

Procedure Init;
 var
  i:integer;
  t:longint;
  offset:longint;
  page  :word;
  size  :word;
 begin
  if not SbSettings then halt;
  if not SbReset then halt;

 { Keys:=0; }
  FillChar(Voices,SizeOf(Voices),0);
  with Voices[0] do begin Noise:=$7FFF; Source:=2; end;
  with Voices[1] do begin Noise:=$7FFF; Source:=0; end;
  with Voices[2] do begin Noise:=$7FFF; Source:=1; end;
(*
  GetDMAMem(Buf,BufSize);  { DMA double-buffer }
  FillChar(Buf^,BufSize,0);

  Dbl:=ptr(Seg(Buf^),Ofs(Buf^)+SndSize);
  Double:=False;

 { IRQ }
  GetIntVec(IRQ+8,OldIRQ);
  SetIntVec(IRQ+8,@SbIRQ8);

 { DMA }
  SetupDMA(DMA,$58,Buf,BufSize-1);

 { DSP - Sets sample rate through internal I/O transfer timer. }
  SBWriteDSP($40);
  SBWriteDSP(SampleTime);
*)
 { Enables speaker output. }
  SBWriteDSP($D1);
(*
 { DSP - Set DMA transfer size }
  if HiSpeed then SBWriteDSP($48) else SBWriteDSP($14);
  SBWriteDSP(lo(SndSize-1));
  SBWriteDSP(hi(SndSize-1));
  if HiSpeed then SBWriteDSP($91);

  EnableIRQ(IRQ);
*)
 end;

Procedure Done;
 begin
(*
  DisableIRQ(IRQ);
  SetIntVec(IRQ+8,OldIRQ);
*)
  SBWriteDSP($D3);
  SBReset;
(*
  FreeMem(Buf,BufSize);
*)
 end;

Procedure VoiceControl(Voice:word);
 var
  NewGate:boolean;
 begin
  With Voices[Voice] do begin

   NewGate:=odd(SIDPoke);
   Sync:=(SIDPoke and 2)>0;
   Ring:=(SIDPoke and 4)>0;
   Test:=(SIDPoke and 8)>0;
   if Test then begin
    Index:=0;
    Noise:=0;
   end else
    Noise:=$7FFF;

   Wave:=SIDPoke shr 4; if Wave>8 then Wave:=0;

   if NewGate and (Not Gate) then begin
    Action:=ATTACK;
    Alarm :=TIMES[A];
    Play:=True;
   end;

   if Gate and (Not NewGate) then begin
    Action:=RELEASE;
    Alarm :=TIMES[R];
   end;

   Gate:=NewGate;

  end;
 end;

Procedure VoiceAD(Voice:word);
 begin
  With Voices[Voice] do begin
   A:=SIDPoke shr 4;
   D:=SIDPoke and 15;
   if Action=ATTACK then Alarm:=TIMES[A] else
   if Action=DECAY  then Alarm:=TIMES[D];
  end;
 end;

Procedure VoiceSR(Voice:word);
 begin
  With Voices[Voice] do begin
   S:=SIDPoke shr 4;
   R:=SIDPoke and 15;
   Level:=S+16*S;
   if Action=RELEASE then Alarm:=TIMES[R];
  end;
 end;

Procedure SetFilter;
 begin
  Voices[0].Filter:=(SIDPoke and 1)>0;
  Voices[1].Filter:=(SIDPoke and 2)>0;
  Voices[2].Filter:=(SIDPoke and 4)>0;
 end;

Procedure SetVolume;
 begin
  Voices[2].Off:=(SIDPoke and $80)>0;
  MaxVol:=(SIDPoke and 15);
 end;

Procedure PokeSID;
 begin
  case SIDAddr of
{ VOICE 0 }
   0,1: Voices[0].Freq :=MemW[CIOSeg:$D400+0];
   2,3: Voices[0].Pulse:=MemW[CIOSeg:$D400+2] and $FFF;

   4: VoiceControl(0);
   5: VoiceAD(0);
   6: VoiceSR(0);

{ VOICE 1 }
   7,8 : Voices[1].Freq :=MemW[CIOSeg:$D400+7];
   9,10: Voices[1].Pulse:=MemW[CIOSeg:$D400+9] and $FFF;

   11: VoiceControl(1);
   12: VoiceAD(1);
   13: VoiceSR(1);

{ VOICE 2 }
   14,15: Voices[2].Freq :=MemW[CIOSeg:$D400+14];
   16,17: Voices[2].Pulse:=MemW[CIOSeg:$D400+16] and $FFF;

   18: VoiceControl(2);
   19: VoiceAD(2);
   20: VoiceSR(2);

{ GENERAL }
{   21,22: FilterFreq:=MemW[CIOSeg:$D400+21]; }
   23   : SetFilter;
   24   : SetVolume;
  end;
 end;
end.