Unit SID; { SID Emulation, from COME BACK 64 v0.92..0.95 }

{ About SID, from "Le Livre du 64" Benot Michel - BCM, 1984


  54296 (D418) = Volume level (0..15)

  FreqFactor = 17,0284 (for a 0,985248 Mhz C64)
  F(Note)=Freq*FreqFactor;
  F(Note)=F(Note-1)*2^(1/12)

  DO,DO#,RE,RE#,MI,FA,FA#,SOL,SOL#,LA,LA#,SI
  LA=440Hz(*FreqFactor)

  S=54272
  Voice1 : POKE S+0,Lo(F)
           POKE S+1,Hi(F)
  Voice2 : POKE S+7,Lo(F)
           POKE S+8,Hi(F)
  Voice3 : POKE S+14,Lo(F)
           POKE S+15,Hi(F)


  Pulse Width % (PW 0<=P<=100)
  P=INT(PW*40.95)  0<=P<=4096
  S=54252
  Voice1 : POKE S+2,Lo(P)
           POKE S+3,Hi(P)
  Voice2 : POKE S+9,Lo(P)
           POKE S+10,Hi(P)
  Voice3 : POKE S+16,Lo(P)
           POKE S+17,Hi(P)


  Wave Start  Stop
  TRI    17    16
  SAW    33    32
  SQR    65    64
  NSE   129   128
 (undocumented)
  TRI2   81    80
  SAW2   97    96


  ADSR
  Value  A   D/R (ms)
    0    2     6 =3*A
    1    8    24
    2   16    48
    3   24    72
    4   38   114
    5   56   168
    6   68   204
    7   80   240
    8  100   300
    9  250   750
   10  500  1500
   11  800  2400
   12 1000  3000
   13 3000  9000
   14 5000 15000
   15 8000 24000

  AD=A*16+D
  SR=S*16+R
  S=54272
  Voice1 : POKE S+5,AD
           POKE S+6,SR
  Voice2 : POKE S+12,AD
           POKE S+13,SR
  Voice3 : POKE S+19,AD
           POKE S+20,SR


  Samples :

   SAW A5  D8 S5  R9 1/2s  = Violon
   TRI A0  D9 S0  R9 1/60s = Xylophone
   SQR A0  D9 S0  R0 1/60s = Piano
   TRI A10 D8 S10 R9  -    = Flute
   TRI A0  D0 S15 R0  -    = Orgue


  S=54272
  POKE S+4,17 :REM START TRI (ATTACK, DECAY)
  FOR T=0 TO 250:NEXT : REM (SUSTAIN ?)
  POSE S+4,16 :REM STOP TRI (RELEASE ?)

 Vol
  :      D
  :     /\
  :    /  \
  :   /    \_______R
  :  /     S       \
  : /               \
  :/.................\....Time
  A

  Attack start at Current Volume (previous note) !

 Vol
  :        D
  :       /\
  :      /  \    D
  :   D /    \__/\  D
  :  /\/     S A  \/\
  : / A           A  \
  :/..................\.....Time
  A


 What I want to do :

  Each Voice have a Timer incremented by SB IRQ.
  When this Timer exceed Alarm, the Volume increase (or decrease) to a
  precomputed Level. When the volume reach this Level, I compute a new
  Alarm and a new Level (and restart Timer).

    ATTACK: begin
     inc(Timer,DeltaTime);        <--- Each IRQ
     if Timer>Alarm then begin    <--- it's time to change volume
      dec(Timer,Alarm);           <--- consume time
      if Volume<Level then        <--- ATTACK done ?
       inc(Volume,DeltaVol)       <---- no, increase volume
      else begin                  <--- yes :
       Volume:=Level;             <---- set volume level (if overflow)
       Action:=DECAY;             <---- now decrease volume
       Timer :=0;                 <---- reset Timer
       Alarm :=TIMES[D]*3;        <---- Compute new Alarm
       Level :=4*D*MaxVol div 255;<---- Compute Target Volume
      end;
     end;
    end;

}
interface

uses dos,data,dialogs,keyboard;

Var
 ADR,IRQ,DMA,HDMA:word;

Procedure Init(AFreq,ARate,AMode:word);
Procedure Done;
Procedure PokeSID; { called by POKE.ASM  }

var
 SIDMode :byte; { 0=None, 1=8Bits, 4=16Bits }

 Volume2 :byte; { Needed by PEEK.ASM }

var
 speed:longint;
 CycleRate:word;

const
 IRQs:word=0;
 currcycles:longint=0;
 lastcycles:longint=0;

implementation

const
 ATTACK =0;
 DECAY  =1;
 SUSTAIN=2;
 RELEASE=3;

 ADSR:array[0..15] of word=( { duration in ms }
  2,8,16,24,38,56,68,80,100,250,500,800,1000,3000,5000,8000
 );

Type
 PSnd=^TSnd;
 TSnd=record
 case boolean of
  true :(raw8 :array[0..2000] of byte);
  false:(raw16:array[0..1000] of integer);
 end;

 TVoice=record
  WAVE   :byte; { WAVE type (can be a mixed type) }
  Index  :word; { Index in this WAVE }
  Volume :word; { Volume Level }
  A,D,S,R:byte; { ATTACK, DECAY, SUSTAIN, RELEASE level }
  Pitch  :word; { Pitch level }
  Ratio  :word;
  Pulse  :word;
  Action :byte; { A/D/S/R }
  Timer  :word;
  Alarm  :word;
  Level  :word;
 end;

Var
 Freq:word;
 Rate:word;
 MixL:word;
 HiSpeed:boolean;

 OldIRQ:pointer;

 Buf:pointer; { DMA double-buffer }
 BufSize:word;
 Dbl:pointer;

 Snd:PSnd;    { sound buffer }
 SndSize:word;
 Double:boolean;

 WAVES:array[0..15,0..63] of integer;

 TIMES:array[0..15] of word;

 Voices:array[0..2] of TVoice;
 Keys  :byte;

 MaxVol:word;
 DeltaTime:word;
 DeltaVolume:word;

{---DMA---}
Const
 DMAADRS:array[0..7] of byte=($00,$02,$04,$06,$C0,$C4,$C8,$CC);
 DMASIZE:array[0..7] of byte=($01,$03,$05,$07,$C2,$C6,$CA,$CE);
 DMAPAGE:array[0..7] of byte=($87,$83,$81,$82,$00,$8B,$89,$8A);

Var
 MSK:byte; { $0A or $D4   }
 MDE:byte; { $0B or $D6   }
 FLP:byte; { $0C or $D8   }
 DAD:byte; { DMAADRS[DMA] }
 DSZ:byte; { DMASIZE[DMA] }
 DPG:byte; { DMAPAGE[DMA] }

Procedure GetDMAMem(var p:pointer; Size:word);
 var
  dmapage:byte;
  dmasize:longint;
  p2     :pointer;
  size2  :word;
 begin
  GetMem(p,Size); { allocate size bytes }
  if p=nil then halt;

  dmapage:=(seg(p^)+ofs(p^) shr 4) shr 12;
  dmasize:=(seg(p^) shl 4+ofs(p^)) and $FFFF;

  if dmasize+size>$FFFF then begin { page boundary }
   freemem(p,size);
   size2:=$FFFF-dmasize;
   getmem(p2,size2);
   if p2=nil then halt;
   getmem(p,size);
   if p=nil then halt;
   freemem(p2,size2);
  end;

  dmapage:=(seg(p^)+ofs(p^) shr 4) shr 12;
  dmasize:=(seg(p^) shl 4+ofs(p^)) and $FFFF;
  if dmasize+size>$FFFF then begin
   writeln('error in GetDMAMem');
   halt;
  end;

 end;

{---DMA (8bits) ---}

Procedure SetupDMA(DMA,Mode:byte; Buff:pointer; Size:word);
 Var
  Len :record
   case boolean of
   true :(long:longint);
   false:(lo,hi:word);
  end;
  Page:word;
  Ofst:word;
 begin
  if DMA<4 then begin
   MSK:=$0A;
   MDE:=$0B;
   FLP:=$0C;
  end else begin
   MSK:=$D4;
   MDE:=$D6;
   FLP:=$D8;
  end;
  DAD:=DMAADRS[DMA];
  DSZ:=DMASIZE[DMA];
  DPG:=DMAPAGE[DMA];

  len.long:=longint(seg(buff^)) shl 4+ofs(buff^);
  page:=len.hi;
  ofst:=len.lo;


  write(#16' DMA Setup :');
  write(' DMA=',DMA);
  write(' MSK=',HexaB(MSK));
  write(' FLP=',HexaB(FLP));
  write(' ADR=',HexaB(DAD));
  write(' PGE=',HexaB(DPG));
  write(' SIZ=',HexaB(DSZ));
  writeln(' MOD=',HexaB(MDE));

  writeln(#254' PGE=',HexaW(Page),' OFS=',HexaW(ofst),' SIZE=',HexaW(SIZE));

  port[MSK]:=DMA or 4; { disable }

  port[DAD]:=lo(ofst);
  port[DAD]:=hi(ofst);
  port[FLP]:=0;        { reset }
  port[DSZ]:=lo(size);
  port[DSZ]:=hi(size);
  port[DPG]:=lo(page);
  if page>255 then port[DPG+$400]:=hi(page);

  port[MDE]:=DMA or Mode;
  port[MSK]:=DMA;      { enable }
 end;

{---IRQ---}
Procedure EnableIRQ(IRQ:byte);
 begin
  Port[$21]:=(Port[$21] and (255 - ( 1 shl IRQ)));
 end;

Procedure DisableIRQ(IRQ:byte);
 begin
  port[$21]:=port[$21] or (1 shl IRQ);
 end;

{---SoundBlaster---}

Function SBReset:boolean;
 var
  i:integer;
  x:byte;
 begin
  Port[ADR+6]:=1; for i:=0 to 3 do x:=Port[ADR+6];
  Port[ADR+6]:=0;
  SBReset:=True;
  for i:=0 to 100 do
   if (Port[ADR+$E] and $80)>0 then { Status }
    if Port[ADR+$A]=$AA then exit;  { Sign   }
  SBReset:=False;
 end;

function SBReadDSP:integer;
 begin
  repeat until (Port[ADR+$E] and $80)>0; { Status }
  SBReadDSP:=Port[ADR+$A]; { Read }
 end;

Procedure SBWriteDSP(Value:byte);
 begin
  repeat until (Port[ADR+$C] and $80)=0; { Write enable }
  Port[ADR+$C]:=Value; { Write }
 end;

Procedure DrawLevel(Top,Level,Volume:word); assembler;
  asm
   MOV DX,03C4h
   MOV AX,0F02h
   OUT DX,AX

   MOV ES,VGASeg
   MOV DI,VGAOfs
   ADD DI,Top

   MOV AL,5
   MOV CX,Volume
   OR  CX,CX
   JZ  @Skip1
    REP STOSB
   @Skip1:

   MOV CX,Level
   SUB CX,Volume
   JS  @Over
   JZ  @Over
    MOV AL,2
    REP STOSB
    MOV CX,255/4
    SUB CX,Level
    JZ  @Done
    JS  @Done
    JMP @EndOfLine
   @Over:
    MOV CX,255/4
    SUB CX,Volume
    JZ  @Done
    JS  @Done
   @EndOfLine:
    MOV AL,0
    REP STOSB
   @Done:
  end;

var
 i:byte;
 value:integer;

procedure Mix8Voice0;
 begin
  with Voices[0] do begin
   case Action of

    ATTACK: begin
     inc(Timer,DeltaTime);
     if Timer>Alarm then begin
      dec(Timer,Alarm);
      if Volume<Level then
       inc(Volume,DeltaVolume)
      else begin
       Volume:=Level;
       Action:=DECAY;
       Timer :=0;
       Alarm :=TIMES[D]*3;
       Level :=4*S*MaxVol div 255;
      end;
     end;
    end;

    DECAY: begin
     inc(Timer,DeltaTime);
     if Timer>Alarm then begin
      dec(Timer,Alarm);
      if Volume>Level then
       dec(Volume,DeltaVolume)
      else begin
       Volume:=Level;
       Action:=SUSTAIN;
       Timer :=0;
       Alarm :=TIMES[S]*3;
       Level :=0;
      end;
     end;
    end;

   { do nothing while SUSTAIN, just wait for a POKE at $D404 }

    RELEASE: begin
     inc(Timer,DeltaTime);
     if Timer>Alarm then begin
      dec(Timer,Alarm);
      if Volume>Level then
       dec(Volume,DeltaVolume)
      else begin
       Volume:=Level;
       Keys  :=Keys and (255-1);
       Timer :=0;
      end;
     end;
    end;

   end;

   DrawLevel(13+222*90,Level div 4,Volume div 4);

   for i:=0 to MixL-1 do begin
    value:=WAVES[WAVE,Index];
    value:=(value*volume) shr 4+ Snd^.Raw8[i];
    if Value<-128 then value:=-128 else
    if Value>+127 then Value:=+127;
    Snd^.Raw8[i]:=Value;
    Inc(Index,Ratio);
    while Index>=64 do dec(Index,64);
   end;
  end;
 end;

procedure Mix8Voice1;
 begin
  with Voices[1] do begin
   case Action of

    ATTACK: begin
     inc(Timer,DeltaTime);
     if Timer>Alarm then begin
      dec(Timer,Alarm);
      if Volume<Level then
       inc(Volume,DeltaVolume)
      else begin
       Volume:=Level;
       Action:=DECAY;
       Timer :=0;
       Alarm :=TIMES[D]*3;
       Level :=4*D*MaxVol div 255;
      end;
     end;
    end;

    DECAY: begin
     inc(Timer,DeltaTime);
     if Timer>Alarm then begin
      dec(Timer,Alarm);
      if Volume>Level then
       dec(Volume,DeltaVolume)
      else begin
       Volume:=Level;
       Action:=SUSTAIN;
       Timer :=0;
       Alarm :=TIMES[S]*3;
       Level :=0;
      end;
     end;
    end;

   { do nothing while SUSTAIN, just wait for a POKE at $D404 }

    RELEASE: begin
     inc(Timer,DeltaTime);
     if Timer>Alarm then begin
      dec(Timer,Alarm);
      if Volume>Level then
       dec(Volume,DeltaVolume)
      else begin
       Volume:=Level;
       Keys  :=Keys and (255-2);
       Timer :=0;
      end;
     end;
    end;

   end;

   DrawLevel(13+224*90,Level div 4,Volume div 4);

   for i:=0 to MixL-1 do begin
    value:=WAVES[WAVE,Index];
    value:=(value*volume) shr 4+ Snd^.Raw8[i];
    if Value<-128 then value:=-128 else
    if Value>+127 then Value:=+127;
    Snd^.Raw8[i]:=Value;
    Inc(Index,Ratio);
    while Index>=64 do dec(Index,64);
   end;
  end;
 end;

procedure Mix8Voice2;
 begin
  with Voices[2] do begin
   case Action of

    ATTACK: begin
     inc(Timer,DeltaTime);
     if Timer>Alarm then begin
      dec(Timer,Alarm);
      if Volume<Level then
       inc(Volume,DeltaVolume)
      else begin
       Volume:=Level;
       Action:=DECAY;
       Timer :=0;
       Alarm :=TIMES[D]*3;
       Level :=4*D*MaxVol div 255;
      end;
     end;
    end;

    DECAY: begin
     inc(Timer,DeltaTime);
     if Timer>Alarm then begin
      dec(Timer,Alarm);
      if Volume>Level then
       dec(Volume,DeltaVolume)
      else begin
       Volume:=Level;
       Action:=SUSTAIN;
       Timer :=0;
       Alarm :=TIMES[S]*3;
       Level :=0;
      end;
     end;
    end;

   { do nothing while SUSTAIN, just wait for a POKE at $D404 }

    RELEASE: begin
     inc(Timer,DeltaTime);
     if Timer>Alarm then begin
      dec(Timer,Alarm);
      if Volume>Level then
       dec(Volume,DeltaVolume)
      else begin
       Volume:=Level;
       Keys  :=Keys and (255-4);
       Timer :=0;
      end;
     end;
    end;

   end;

   DrawLevel(13+226*90,Level div 4,Volume div 4);

   for i:=0 to MixL-1 do begin
    value:=WAVES[WAVE,Index];
    value:=(value*volume) shr 4+ Snd^.Raw8[i];
    if Value<-128 then value:=-128 else
    if Value>+127 then Value:=+127;
    Snd^.Raw8[i]:=Value;
    Inc(Index,Ratio);
    while Index>=64 do dec(Index,64);
   end;
  end;
 end;

procedure SbIRQ8; interrupt;
 var
  dummy:word;
 begin
  inc(IRQs);

  if double then begin
   Snd:=Dbl;
   double:=false;
  end else begin
   Snd:=Buf;
   double:=true;
  end;

  asm
   mov cx,SndSize
   shr cx,1
   les di,Snd
   xor ax,ax
   rep stosw
  end;

  if (Keys and 1)=1 then Mix8Voice0;
  if (Keys and 2)=2 then Mix8Voice1;
  if (Keys and 4)=4 then Mix8Voice2;

  if HiSpeed then SBWriteDSP($48) else SBWriteDSP($14);
  SBWriteDSP(lo(MixL-1));
  SBWriteDSP(hi(MixL-1));
  if HiSpeed then SBWriteDSP($91);

  dummy:=Port[ADR+$E]; { Interruption Ok }
  Port[$20]:=$20;      { End of IRQ }
 end;

{---VIC---}
Procedure InitWaves;
 Var
  TRI,SAW,SQR,NSE:array[0..63] of shortint;
  i,j:integer;
  t  :integer;
 begin
  t:=-128;

  for i:=0 to 31 do begin
   TRI[i]:=t;
   SQR[i]:=-128;
   inc(t,4);
  end;

  for i:=32 to 63 do begin
   dec(t,4);
   TRI[i]:=t;
   SQR[i]:=+127;
  end;

  for i:=0 to 63 do begin
   SAW[i]:=t;
   NSE[i]:=random(255)-128;
   inc(t,4);
  end;

  for j:=1 to 16 do begin
   for i:=0 to 63 do begin
    t:=-128;
    if (j and 1)>0 then inc(t,TRI[i]);
    if (j and 2)>0 then inc(t,SAW[i]);
    if (j and 4)>0 then inc(t,SQR[i]);
    if (j and 8)>0 then inc(t,NSE[i]);
    if t<-128 then t:=-128 else
    if t>+127 then t:=+127;
    WAVES[j-1][i]:=t;
   end;
  end;
 end;

Function SbSettings:boolean;
 var
  Blaster:string;
  p:byte;
 begin
  Write(#16' Init Sound Card - ');

  Blaster:=GetEnv('BLASTER'); { BLASTER=A220 I5 D1 H5 }
  SbSettings:=False;
  p:=pos('A',Blaster); if p=0 then exit;
  ADR:=$200+16*(ord(Blaster[p+2])-ord('0')); { A2x0 }
  p:=pos('I',Blaster); if p=0 then exit;
  IRQ:=ord(Blaster[p+1])-ord('0');            { I5 }
  p:=pos('D',Blaster); if p=0 then exit;
  DMA:=ord(Blaster[p+1])-ord('0');            { D1 }

  SbSettings:=True;

  p:=pos('H',Blaster);
  if p=0 then HDMA:=0 else HDMA:=ord(Blaster[p+1])-ord('0');  { H5 }

  Blaster:='BLASTER='+HexaW(ADR);
  Blaster[9]:='A';

  WriteLn(Blaster,' I',IRQ,' D',DMA,' H',HDMA);
 end;

Procedure Init(AFreq,ARate,AMode:word);
 var
  i:integer;
  t:longint;
  offset:longint;
  page  :word;
  size  :word;
  Time  :byte;
 begin
  if not SbSettings then halt;
  if not SbReset then halt;

  InitWaves; { compute C64 Waves }

  Keys:=0;
  Volume2:=0;
  FillChar(Voices,SizeOf(Voices),0);

  Freq:=AFreq;
  Rate:=ARate;
  MixL:=Freq div Rate;     { mixing buffer length }

  if AMode>0 then
   SIDMode:=AMode
  else begin
   SBWriteDSP($E1);
   SIDMode:=SBReadDSP;
   SBReadDSP;
   if SIDMode<4 then SIDMode:=1; { 8 bits }
  end;

  CycleRate:=9867 div Rate;{ PAL }

  if SIDMode=1 then
   SndSize:=MixL        { array of byte : 8 bits DMA }
  else
   SndSize:=MixL shl 1; { array of integer : 16 bits DMA }

  BufSize:=SndSize shl 1;
  GetDMAMem(Buf,BufSize);  { mixing double-buffer }
  FillChar(Buf^,BufSize,0);

  Dbl:=ptr(Seg(Buf^),Ofs(Buf^)+SndSize);

  Double:=False;

  for i:=0 to 15 do TIMES[i]:=ADSR[i];

  DeltaTime:=CycleRate;

  DeltaVolume:=1;

 { IRQ }
  GetIntVec(IRQ+8,OldIRQ);

 { DMA }
{  if SIDMode=1 then begin { 8bits }

   SetIntVec(IRQ+8,@SbIRQ8);
   SetupDMA(DMA,$58,Buf,BufSize-1);

   Readln;

  { DSP - Sets sample rate through internal I/O transfer timer. }
   Time:=256 - (1000000 div Freq);
   SBWriteDSP($40);
   SBWriteDSP(Time);

   HiSpeed:=(1000000 div (256-Time))>22222;

  { Enables speaker output. }
   SBWriteDSP($D1);

  { DSP - Set DMA transfer size }
   if HiSpeed then SBWriteDSP($48) else SBWriteDSP($14);
   SBWriteDSP(lo(MixL-1));
   SBWriteDSP(hi(MixL-1));
   if HiSpeed then SBWriteDSP($91);

(*
  end else begin { 16bits }
   SetIntVec(IRQ+8,@SbIRQ16);
   SetupDMA(HDMA,$58,Buf,SndSize-1);
   SBWriteDSP($41);
   SBWriteDSP(hi(Freq));
   SBWriteDSP(lo(Freq));
   SBWriteDSP($B6);
   SBWriteDSP($10);
   SBWriteDSP(Lo(MixL-1));
   SBWriteDSP(Hi(MixL-1));
   SBWriteDSP($D1);
  end;
*)

  EnableIRQ(IRQ);

 end;

Procedure Done;
 begin
  DisableIRQ(IRQ);
  SetIntVec(IRQ+8,OldIRQ);
  if SIDMode=1 then begin
   SBWriteDSP($D3);
   SBReset;
  end else begin
   SBWriteDSP($D5);
  end;
  FreeMem(Buf,BufSize);
 end;

Procedure PokeSID;
 begin
  case SIDAddr of
{ VOICE 0 }
   0,1: with Voices[0] do begin
         Pitch:=MemW[CIOSeg:$D400];
         Ratio:=longint(Pitch shl 8) div Freq+1;
        end;

   2,3: Voices[0].Pulse:=MemW[CIOSeg:$D402];

   4: with Voices[0] do begin
        Wave  :=Hi(SIDPoke);
        Index :=0;
        if odd(SIDPoke) then begin
         Action:=ATTACK;
         Timer :=0;
         Alarm :=TIMES[A];
         Level :=MaxVol;
         Keys  :=Keys or 1;
        end else begin
         Action:=RELEASE;
         Timer :=0;
         Alarm :=TIMES[R]*3;
         Level :=0;
        end;
      end;

   5: begin
       Voices[0].A:=SIDPoke shr 4;
       Voices[0].D:=SIDPoke and 15;
      end;
   6: begin
       Voices[0].S:=SIDPoke shr 4;
       Voices[0].R:=SIDPoke and 15;
      end;
{ VOICE 1 }
   7,8: with Voices[1] do begin
         Pitch:=MemW[CIOSeg:$D400+7];
         Ratio:=longint(Pitch shl 16) div Freq+1;
        end;

   9,10: Voices[1].Pulse:=MemW[CIOSeg:$D400+9];

   11: with Voices[1] do begin
        Wave  :=Hi(SIDPoke);
        Index :=0;
        if odd(SIDPoke) then begin
         Action:=ATTACK;
         Timer :=0;
         Alarm :=TIMES[A];
         Level :=MaxVol;
         Keys  :=Keys or 2;
        end else begin
         Action:=RELEASE;
         Timer :=0;
         Alarm :=TIMES[R]*3;
         Level :=0;
        end;
      end;

   12: begin
       Voices[1].A:=SIDPoke shr 4;
       Voices[1].D:=SIDPoke and 15;
      end;
   13: begin
       Voices[1].S:=SIDPoke shr 4;
       Voices[1].R:=SIDPoke and 15;
      end;
{ VOICE 2 }
   14,15: with Voices[2] do begin
         Pitch:=MemW[CIOSeg:$D400+14];
         Ratio:=longint(Pitch shl 16) div Freq+1;
        end;

   16,17: Voices[2].Pulse:=MemW[CIOSeg:$D400+16];

   18: with Voices[2] do begin
        Wave  :=Hi(SIDPoke);
        Index :=0;
        if odd(SIDPoke) then begin
         Action:=ATTACK;
         Timer :=0;
         Alarm :=TIMES[A];
         Level :=MaxVol;
         Keys  :=Keys or 4;
        end else begin
         Action:=RELEASE;
         Timer :=0;
         Alarm :=TIMES[R]*3;
         Level :=0;
        end;
      end;

   19: begin
       Voices[2].A:=SIDPoke shr 4;
       Voices[2].D:=SIDPoke and 15;
      end;
   20: begin
       Voices[2].S:=SIDPoke shr 4;
       Voices[2].R:=SIDPoke and 15;
      end;
   24: MaxVol:=(SIDPoke and 15)*17;
(*

   7,8: with SndVoices[1] do begin
         Pitch:=MemW[CIOSeg:$D400+7] shl 2;
         ratio:=(pitch shl Shift) div Freq;
        end;

   11: with SndVoices[1] do begin
       if ((SIDPoke xor SIDValue) and $F1)>0 then begin
        Samp:=SIDPoke shr 4;
        SPtr:=0;
        envx:=0;
        envstate:=ATTACK;
        SndKeys:=SndKeys or 2;
       end else begin
        EnvState:=RELEASE;
       end;
      end;

   12:begin
       SndVoices[1].A:=SIDValue shr 4;
       SndVoices[1].D:=SIDValue and 15;
      end;
   13:begin
       SndVoices[1].S:=SIDValue shr 4;
       SndVoices[1].R:=SIDValue and 15;
      end;


   14,15: with SndVoices[2] do begin
           Pitch:=MemW[CIOSeg:$D400+14] shl 2;
           ratio:=(pitch shl Shift) div Freq;
          end;

   18: with SndVoices[2] do begin
       if ((SIDPoke xor SIDValue) and $F1)>0 then begin
        Samp:=SIDPoke shr 4;
        SPtr:=0;
        envx:=0;
        envstate:=ATTACK;
        SndKeys:=SndKeys or 4;
       end else begin
        EnvState:=RELEASE;
       end;
      end;

   19:begin
       SndVoices[2].A:=SIDValue shr 4;
       SndVoices[2].D:=SIDValue and 15;
      end;
   20:begin
       SndVoices[2].S:=SIDValue shr 4;
       SndVoices[2].R:=SIDValue and 15;
      end;
*)
  end;
 end;
end.