Unit SID8; { SID Emulation, from COME BACK 64 v0.92..0.95 }

{ About SID, from "Le Livre du 64" Benot Michel - BCM, 1984


  54296 (D418) = Volume level (0..15)

  FreqFactor = 17,0284 (for a 0,985248 Mhz C64)
  F(Note)=Freq*FreqFactor;
  F(Note)=F(Note-1)*2^(1/12)

  DO,DO#,RE,RE#,MI,FA,FA#,SOL,SOL#,LA,LA#,SI
  LA=440Hz(*FreqFactor)

  S=54272
  Voice1 : POKE S+0,Lo(F)
           POKE S+1,Hi(F)
  Voice2 : POKE S+7,Lo(F)
           POKE S+8,Hi(F)
  Voice3 : POKE S+14,Lo(F)
           POKE S+15,Hi(F)


  Pulse Width % (PW 0<=P<=100)
  P=INT(PW*40.95)  0<=P<=4096
  S=54252
  Voice1 : POKE S+2,Lo(P)
           POKE S+3,Hi(P)
  Voice2 : POKE S+9,Lo(P)
           POKE S+10,Hi(P)
  Voice3 : POKE S+16,Lo(P)
           POKE S+17,Hi(P)


  Wave Start  Stop
  TRI    17    16      1-0000 1
  SAW    33    32     10-0000 2
  SQR    65    64    100-0000 4
  NSE   129   128   1000-0000 8
 (undocumented)
  TRI2   81    80
  SAW2   97    96


  ADSR
  Value  A   D/R (ms)
    0    2     6 =3*A
    1    8    24
    2   16    48
    3   24    72
    4   38   114
    5   56   168
    6   68   204
    7   80   240
    8  100   300
    9  250   750
   10  500  1500
   11  800  2400
   12 1000  3000
   13 3000  9000
   14 5000 15000
   15 8000 24000

  AD=A*16+D
  SR=S*16+R
  S=54272
  Voice1 : POKE S+5,AD
           POKE S+6,SR
  Voice2 : POKE S+12,AD
           POKE S+13,SR
  Voice3 : POKE S+19,AD
           POKE S+20,SR


  Samples :

   SAW A5  D8 S5  R9 1/2s  = Violon
   TRI A0  D9 S0  R9 1/60s = Xylophone
   SQR A0  D9 S0  R0 1/60s = Piano
   TRI A10 D8 S10 R9  -    = Flute
   TRI A0  D0 S15 R0  -    = Orgue


  S=54272
  POKE S+4,17 :REM START TRI (ATTACK, DECAY)
  FOR T=0 TO 250:NEXT : REM (SUSTAIN ?)
  POSE S+4,16 :REM STOP TRI (RELEASE ?)

 Vol
  :      D
  :     /\
  :    /  \
  :   /    \_______R
  :  /     S       \
  : /               \
  :/.................\....Time
  A

  Attack start at Current Volume (previous note) !

 Vol
  :        D
  :       /\
  :      /  \    D
  :   D /    \__/\  D
  :  /\/     S A  \/\
  : / A           A  \
  :/..................\.....Time
  A


 What I want to do :

  Each Voice have a Timer incremented by SB IRQ.
  When this Timer exceed Alarm, the Volume increase (or decrease) to a
  precomputed Level. When the volume reach this Level, I compute a new
  Alarm and a new Level (and restart Timer).

    ATTACK: begin
     inc(Timer,DeltaTime);        <--- Each IRQ
     if Timer>Alarm then begin    <--- it's time to change volume
      dec(Timer,Alarm);           <--- consume time
      if Volume<Level then        <--- ATTACK done ?
       inc(Volume,DeltaVol)       <---- no, increase volume
      else begin                  <--- yes :
       Volume:=Level;             <---- set volume level (if overflow)
       Action:=DECAY;             <---- now decrease volume
       Timer :=0;                 <---- reset Timer
       Alarm :=TIMES[D]*3;        <---- Compute new Alarm
       Level :=4*D*MaxVol div 255;<---- Compute Target Volume
      end;
     end;
    end;

 SAMPLE PROGRAM :

  10 VOL = 15      : REM SOUND VOLUME
  20 FL=0:FH=20    : REM SOUND FREQ (LOW AND HIGH)
  30 AD=0:SR=16*15 : REM A=0 D=0 S=15 R=0
  40 WAV=16        : REM TRIANGLE WAVE
  110 POKE 54272+24,VOL   : REM 54272=$D400
  120 POKE 54272+0,FL
  130 POKE 54272+1,FH
  140 POKE 54272+5,AD
  150 POKE 54272+6,SR
  160 POKE 54272+4,WAV+1  : REM START SOUND
  170 FOR T=0 TO 500:NEXT : REM SOUND DURATION
  180 POKE 54272+4,WAV    : REM STOP SOUND

}
interface

uses dos,data,dialogs,keyboard;

Const
 SampleRate = 22000; { MHz }
 SampleTime = 256 - (1000000 div SampleRate);
 HiSpeed:boolean=(1000000 div (256-SampleTime))>22222;

 ADSR2TIME  = 5;

 SndSize    = ADSR2TIME*64;     { sampling buffer size }
 BufSize    = 2*SndSize; { DMA ring-buffer size }

 VolScale   = 1; { MaxVol }
 VolConst   = 15*VolScale;

 WaveScale  = 1; { InitWaves }
 WaveLo     = -128*WaveScale;
 WaveHi     = +127*WaveScale;
 WaveRange  = (WaveHi-WaveLo);
 WaveStep   = WaveRange div 64;

 DeltaTime  = 1; { Mix8Voice }
 DeltaVolume= 1;

Var
 ADR,IRQ,DMA:word; { BLASTER=Axxx Ix Dx }

Procedure Init;
Procedure Done;

Procedure PokeSID; { called by POKE.ASM  }

var
 SIDAddr :word;
 SIDValue:byte;
 SIDPoke :byte;

 Volume2 :byte; { Needed by PEEK.ASM }

Var
 speed:longint;
 CycleRate:word;

const
 IRQs:word=0;
 currcycles:longint=0;
 lastcycles:longint=0;

Var
 BIOSTicks:longint absolute $40:$6C;
 SBStart:longint;

implementation

const
 ATTACK =0;
 DECAY  =1;
 SUSTAIN=2;
 RELEASE=3;

 ADSR:array[0..15] of word=( { duration in ms }
  2,8,16,24,38,56,68,80,100,250,500,800,1000,3000,5000,8000
 );

Type
 PSnd=^TSnd;
 TSnd=array[0..SndSize-1] of byte;

Var
 OldIRQ:pointer;

 Buf:pointer; { DMA double-buffer }
 Dbl:pointer; { double buffer adr }
 Snd:PSnd;    { sound buffer }
 Double:boolean;

 WAVES:array[1..8,0..63] of integer;

 TIMES:array[0..15] of word;

{---DMA---}
Const
 DMAADRS:array[0..7] of byte=($00,$02,$04,$06,$C0,$C4,$C8,$CC);
 DMASIZE:array[0..7] of byte=($01,$03,$05,$07,$C2,$C6,$CA,$CE);
 DMAPAGE:array[0..7] of byte=($87,$83,$81,$82,$00,$8B,$89,$8A);

Var
 MSK:byte; { $0A or $D4   }
 MDE:byte; { $0B or $D6   }
 FLP:byte; { $0C or $D8   }
 DAD:byte; { DMAADRS[DMA] }
 DSZ:byte; { DMASIZE[DMA] }
 DPG:byte; { DMAPAGE[DMA] }

Procedure GetDMAMem(var p:pointer; Size:word);
 var
  dmapage:byte;
  dmasize:longint;
  p2     :pointer;
  size2  :word;
 begin
  GetMem(p,Size); { allocate size bytes }
  if p=nil then halt;

  dmapage:=(seg(p^)+ofs(p^) shr 4) shr 12;
  dmasize:=(seg(p^) shl 4+ofs(p^)) and $FFFF;

  if dmasize+size>$FFFF then begin { page boundary }
   freemem(p,size);
   size2:=$FFFF-dmasize;
   getmem(p2,size2);
   if p2=nil then halt;
   getmem(p,size);
   if p=nil then halt;
   freemem(p2,size2);
  end;
 end;

{---DMA (8bits) ---}

Procedure SetupDMA(DMA,Mode:byte; Buff:pointer; Size:word);
 Var
  Len :record
   case boolean of
   true :(long:longint);
   false:(lo,hi:word);
  end;
  Page:word;
  Ofst:word;
 begin
  if DMA<4 then begin
   MSK:=$0A;
   MDE:=$0B;
   FLP:=$0C;
  end else begin
   MSK:=$D4;
   MDE:=$D6;
   FLP:=$D8;
  end;
  DAD:=DMAADRS[DMA];
  DSZ:=DMASIZE[DMA];
  DPG:=DMAPAGE[DMA];

  len.long:=longint(seg(buff^)) shl 4+ofs(buff^);
  page:=len.hi;
  ofst:=len.lo;

  write(#16' DMA Setup :');
  write(' DMA=',DMA);
  write(' MSK=',HexaB(MSK));
  write(' FLP=',HexaB(FLP));
  write(' ADR=',HexaB(DAD));
  write(' PGE=',HexaB(DPG));
  write(' SIZ=',HexaB(DSZ));
  writeln(' MOD=',HexaB(MDE));

  writeln(#254' PGE=',HexaW(Page),' OFS=',HexaW(ofst),' SIZE=',HexaW(SIZE));

  DMA:=DMA and 3;

  port[MSK]:=DMA or 4; { disable }

  port[DAD]:=lo(ofst);
  port[DAD]:=hi(ofst);
  port[FLP]:=0;        { reset }
  port[DSZ]:=lo(size);
  port[DSZ]:=hi(size);
  port[DPG]:=lo(page);
  if page>255 then port[DPG+$400]:=hi(page);

  port[MDE]:=DMA or Mode;
  port[MSK]:=DMA;      { enable }
 end;

{---IRQ---}
Procedure EnableIRQ(IRQ:byte);
 begin
  Port[$21]:=(Port[$21] and (255 - ( 1 shl IRQ)));
 end;

Procedure DisableIRQ(IRQ:byte);
 begin
  port[$21]:=port[$21] or (1 shl IRQ);
 end;

{---SoundBlaster---}

Function SBReset:boolean;
 var
  i:integer;
  x:byte;
 begin
  Port[ADR+6]:=1; for i:=0 to 3 do x:=Port[ADR+6];
  Port[ADR+6]:=0;
  SBReset:=True;
  for i:=0 to 100 do
   if (Port[ADR+$E] and $80)>0 then { Status }
    if Port[ADR+$A]=$AA then exit;  { Sign   }
  SBReset:=False;
 end;

function SBReadDSP:integer;
 begin
  repeat until (Port[ADR+$E] and $80)>0; { Status }
  SBReadDSP:=Port[ADR+$A]; { Read }
 end;

Procedure SBWriteDSP(Value:byte);
 begin
  repeat until (Port[ADR+$C] and $80)=0; { Write enable }
  Port[ADR+$C]:=Value; { Write }
 end;

Type
 TVoice=record
  WAVE   :word;    { WAVE type (can be a mixed type) }
  Index  :word;    { Index in this WAVE }
  Volume :integer; { Volume Level }
  A,D,S,R:word;    { ATTACK, DECAY, SUSTAIN, RELEASE level }
  Pitch  :word;    { Pitch level }
  Ratio  :word;
  Pulse  :word;
  Action :word;    { A/D/S/R }
  Timer  :word;
  Alarm  :word;
  Level  :integer;
 end;

var
 MaxVol  :word;
 Voices:array[0..2] of TVoice;
 Keys  :byte;
 Mix:array[0..SndSize-1] of integer;

 i:integer;

procedure MixVoice(Voice:word);
 begin
  with Voices[Voice] do begin         { with this voice }

   for i:=0 to SndSize-1 do begin     { and for all the mixing buffer }

   case Action of                     { case Wave state }

    ATTACK: begin                     { ATTACK : }
     if Volume<Level then begin       { Until target level }
      inc(Timer,DeltaTime);           { count time }
      while Timer>=Alarm do begin     { enough time ? }
       dec(Timer,Alarm);              { consume time  }
       inc(Volume,DeltaVolume)        { change volume }
      end;
     end;
     if Volume>=Level then begin      { target volume ? }
      Volume:=Level;                  { set it }
      Action:=DECAY;                  { new action }
      Timer :=0;                      { reset timer }
      Level :=S*MaxVol div VolConst;  { reset target volume }
      Alarm:=TIMES[D]*3               { time between change }
     end;
    end;

    DECAY: begin
     if Volume>Level then begin
      inc(Timer,DeltaTime);
      while Timer>=Alarm do begin
       dec(Timer,Alarm);
       dec(Volume,DeltaVolume)
      end;
     end;
     if Volume<=Level then begin
      Volume:=Level;
      Action:=SUSTAIN;
     end;
    end;

   { do nothing while SUSTAIN, just wait for a POKE at $D400+7*Voice+4 }

    RELEASE: begin
     if Volume>Level then begin
      inc(Timer,DeltaTime);
      while Timer>=Alarm do begin
       dec(Timer,Alarm);
       dec(Volume,DeltaVolume)
      end;
     end;
     if Volume<=Level then begin
      Volume:=Level;
      Keys  :=Keys and (255-1 shl Voice);
      Timer :=0;
     end;
    end;

   end;

    Inc(Mix[i],WAVES[Wave,Index] * Volume div VolConst);

    Index :=(Index+Ratio) and 63;
   end;

  end;
 end;

Var
 CPU:byte;

procedure SbIRQ8; interrupt;
 var
  i,t:integer;
 begin
  CPU:=Test8086;
  Test8086:=0;   { BP7 bug on Longint math (use 32bits regs not saved) }

  inc(IRQs);

  if double then begin
   Snd:=Dbl;
   double:=false;
  end else begin
   Snd:=Buf;
   double:=true;
  end;

  if Keys=0 then begin

   FillChar(Snd^,SndSize,0);

  end else begin

   FillChar(Mix,SizeOf(Mix),0); { 16bits sample }

   if (Keys and 1)=1 then MixVoice(0);
   if (Keys and 2)=2 then MixVoice(1);
   if (Keys and 4)=4 then MixVoice(2);

   for i:=0 to SndSize-1 do begin
    t:=Mix[i] div WaveScale;
    if t<-128 then t:=-128 else
    if t>+127 then t:=+127;
    Snd^[i]:=t+128;
   end;

  end;

  if HiSpeed then SBWriteDSP($48) else SBWriteDSP($14);
  SBWriteDSP(lo(SndSize-1));
  SBWriteDSP(hi(SndSize-1));
  if HiSpeed then SBWriteDSP($91);

  Test8086:=CPU;

  i:=Port[ADR+$E]; { Interruption Ok }
  Port[$20]:=$20;      { End of IRQ }
 end;

{---VIC---}
Procedure InitWaves;
 Var
  TRI,SAW,SQR,NSE:array[0..63] of integer;
  i,j:integer;
  t  :integer;
 begin
  t:=WaveLo;

  for i:=0 to 31 do begin
   inc(t,WaveStep);
   TRI[i]:=t;
   SQR[i]:=WaveLo;
  end;

  for i:=32 to 63 do begin
   dec(t,WaveStep);
   TRI[i]:=t;
   SQR[i]:=WaveHi;
  end;

  for i:=0 to 63 do begin
   inc(t,WaveStep);
   SAW[i]:=t;
   NSE[i]:=random(WaveRange)+WaveLo;
  end;

  for j:=1 to 8 do begin { mix waves }
   for i:=0 to 63 do begin
    t:=0;
    if (j and 1)>0 then inc(t,TRI[i]);
    if (j and 2)>0 then inc(t,SAW[i]);
    if (j and 4)>0 then inc(t,SQR[i]);
    if (j and 8)>0 then inc(t,NSE[i]);
   (*
    if t<WaveLo then t:=WaveLo else
    if t>WaveHi then t:=WaveHi;
   *)
    WAVES[j][i]:=t;
   end;
  end;

 end;

Function SbSettings:boolean;
 var
  Blaster:string;
  p:byte;
 begin
  Write(#16' Init Sound Card - ');

  Blaster:=GetEnv('BLASTER'); { BLASTER=A220 I5 D1 H5 }
  SbSettings:=False;
  p:=pos('A',Blaster); if p=0 then exit;
  ADR:=$200+16*(ord(Blaster[p+2])-ord('0')); { A2x0 }
  p:=pos('I',Blaster); if p=0 then exit;
  IRQ:=ord(Blaster[p+1])-ord('0');            { I5 }
  p:=pos('D',Blaster); if p=0 then exit;
  DMA:=ord(Blaster[p+1])-ord('0');            { D1 }

  SbSettings:=True;

  Blaster:='BLASTER='+HexaW(ADR);
  Blaster[9]:='A';

  WriteLn(Blaster,' I',IRQ,' D',DMA);
 end;

Procedure Init;
 var
  i:integer;
  t:longint;
  offset:longint;
  page  :word;
  size  :word;
 begin
  if not SbSettings then halt;
  if not SbReset then halt;

  InitWaves; { compute C64 8bits Waves }

  Keys:=0;
  Volume2:=0;
  FillChar(Voices,SizeOf(Voices),0);

  CycleRate:=9867 div (SampleRate div SndSize); { PAL }

  GetDMAMem(Buf,BufSize);  { DMA double-buffer }
  FillChar(Buf^,BufSize,0);

  Dbl:=ptr(Seg(Buf^),Ofs(Buf^)+SndSize);
  Double:=False;

  for i:=0 to 15 do TIMES[i]:=ADSR[i]*ADSR2TIME; { ??? }

 { IRQ }
  GetIntVec(IRQ+8,OldIRQ);

 { DMA }
  SetIntVec(IRQ+8,@SbIRQ8);
  SetupDMA(DMA,$58,Buf,BufSize-1);

  Readln;

 { DSP - Sets sample rate through internal I/O transfer timer. }
  SBWriteDSP($40);
  SBWriteDSP(SampleTime);

 { Enables speaker output. }
  SBWriteDSP($D1);

 { DSP - Set DMA transfer size }
  if HiSpeed then SBWriteDSP($48) else SBWriteDSP($14);
  SBWriteDSP(lo(SndSize-1));
  SBWriteDSP(hi(SndSize-1));
  if HiSpeed then SBWriteDSP($91);

  SBStart:=BIOSTicks;

  EnableIRQ(IRQ);
 end;

Procedure Done;
 begin
  DisableIRQ(IRQ);
  SetIntVec(IRQ+8,OldIRQ);
  SBWriteDSP($D3);
  SBReset;
  FreeMem(Buf,BufSize);
 end;

Function PitchToRatio(Pitch:word):word;
 begin
  PitchToRatio:={(Pitch shl PitchShift) div (SampleRate shr SampleShift)+1;}
   Pitch div (64*(SampleRate) div 1000)+1;
 end;

Procedure VoiceControl(Voice:word);
 begin
  With Voices[Voice] do begin
   Wave :=SIDPoke shr 4;
   if Wave>8 then Wave:=0;
   Index:=0;
   if (Wave>0) and odd(SIDPoke) then begin
    Action:=ATTACK;
    Level :=MaxVol;
    Timer :=0;
    Alarm :=TIMES[A];
    Keys  :=Keys or (1 shl Voice);
   end else begin
    Action:=RELEASE;
    Level :=0;
    Timer :=0;
    Alarm :=TIMES[R]*3;
   end;
  end;
 end;

Procedure PokeSID;
 begin
  case SIDAddr of
{ VOICE 0 }
   0,1: with Voices[0] do begin
         Pitch:=MemW[CIOSeg:$D400];
         Ratio:=PitchToRatio(Pitch);
        end;

   2,3: Voices[0].Pulse:=MemW[CIOSeg:$D402];

   4: VoiceControl(0);

   5: begin
       Voices[0].A:=SIDPoke shr 4;
       Voices[0].D:=SIDPoke and 15;
      end;
   6: begin
       Voices[0].S:=SIDPoke shr 4;
       Voices[0].R:=SIDPoke and 15;
      end;
{ VOICE 1 }
   7,8: with Voices[1] do begin
         Pitch:=MemW[CIOSeg:$D400+7];
         Ratio:=PitchToRatio(Pitch);
        end;

   9,10: Voices[1].Pulse:=MemW[CIOSeg:$D400+9];

   11: VoiceControl(1);

   12: begin
       Voices[1].A:=SIDPoke shr 4;
       Voices[1].D:=SIDPoke and 15;
      end;
   13: begin
       Voices[1].S:=SIDPoke shr 4;
       Voices[1].R:=SIDPoke and 15;
      end;
{ VOICE 2 }
   14,15: with Voices[2] do begin
         Pitch:=MemW[CIOSeg:$D400+14];
         Ratio:=PitchToRatio(Pitch);
        end;

   16,17: Voices[2].Pulse:=MemW[CIOSeg:$D400+16];

   18: VoiceControl(2);

   19: begin
       Voices[2].A:=SIDPoke shr 4;
       Voices[2].D:=SIDPoke and 15;
      end;
   20: begin
       Voices[2].S:=SIDPoke shr 4;
       Voices[2].R:=SIDPoke and 15;
      end;
{ GENERAL }
   24: MaxVol:=(SIDPoke and 15)*VolScale{*17};
  end;
 end;
end.